#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.system.env import env
from dev0s.classes.defaults.files import *
from dev0s.classes.defaults import objects
from dev0s.classes import utils
from dev0s.classes import cli
import platform 

# the defaults. class.
class Defaults(object):
	def __init__(self):

		# docs.
		DOCS = {
			"module":"dev0s.defaults", 
			"initialized":True,
			"description":[], 
			"chapter": "Defaults", }

		# vars.
		self.vars = objects.Object({
			"os": platform.system().lower(),
			"user": os.environ.get("USER"),
			"owner": os.environ.get("USER"),
			"group": "root",
			"executable":FilePath(sys.executable),
			"media": Directory(f"/media/{os.environ.get('USER')}"),
			"homes": Directory(gfp.base(path=os.environ.get('HOME'))),
			"home": Directory(gfp.clean(os.environ.get('HOME')+"/")),
			"pwd":Directory(gfp.clean(utils.__execute_script__("pwd").replace("\n",""))),
			"site_packages":self.site_packages(),
		})
		if self.vars.os in ["darwin"]:  
			self.vars.os = "macos"
			self.vars.group = "staff"
			self.vars.media = f"/Volumes/"
			self.vars.python3 = f"/usr/bin/python3"
		if self.vars.site_packages.fp.name() not in ["site-packages", "dist-packages"]: 
			self.vars.site_packages = env.get("SITE_PACKAGES", default=None)
			if self.vars.site_packages == None:
				raise ValueError(f"<dev0s.defaults>: Define environment variable [SITE_PACKAGES] ({ALIAS} developer mode) ({ALIAS} base: {gfp.base(SOURCE_PATH)}).")

		# universal options.
		self.options = objects.Object({
			"log_level": int(cli.get_argument("--log-level", required=False, default=env.get_integer("LOG_LEVEL", default=0))),
			"cli": env.get_boolean("CLI", default=False),
			"json": cli.arguments_present(["-j", "--json"]),
			"interactive": self.interactive(),
			"forced": cli.arguments_present(["--forced", "-f", "--force"], default=env.get("FORCED", default=False, format=bool)),
			"quiet": cli.arguments_present(["--quiet", "-q", "-s", "--silent"], default=env.get("QUIET", default=False, format=bool)),
		})

		# logging.
		if self.options.log_level >= 2:
			print(f"{ALIAS}:")
			print(f"  * vars: ")
			print(f"    - os: {self.vars.os}")
			print(f"    - user: {self.vars.user}")
			print(f"    - owner: {self.vars.owner}")
			print(f"    - group: {self.vars.group}")
			print(f"    - executable: {self.vars.executable}")
			print(f"    - media: {self.vars.media}")
			print(f"    - homes: {self.vars.homes}")
			print(f"    - home: {self.vars.home}")
			print(f"    - pwd: {self.vars.pwd}")
			print(f"    - site_packages: {self.vars.site_packages}")
			print(f"  * options: ")
			print(f"    - log_level: {self.options.log_level}")
			print(f"    - cli: {self.options.cli}")
			print(f"    - json: {self.options.json}")
			print(f"    - interactive: {self.options.interactive}")
			print(f"    - forced: {self.options.forced}")
			print(f"    - quiet: {self.options.quiet}")

	# get operating system.
	def operating_system(self, supported=["*"]):
		os = self.vars.os
		if os in ["macos"] and ("*" in supported or os in supported): return "macos"
		elif os in ["linux"] and ("*" in supported or os in supported): return "linux"
		else: raise ValueError(f"Unsupported operating system: [{os}].")

	# create alias.
	def alias(self,
		# the source name.
		alias=None, 
		# the source path.
		executable=None,
		# can use sudo.
		sudo=False,
		# overwrite.
		overwrite=False,
		# the venv path (leave None to ignore).
		venv=None,
		# the path to the requirements file (leave None to ignore).
		requirements=None,
	):
		l_alias = cli.get_argument("--create-alias", required=False)
		present = "--create-alias" in sys.argv and l_alias == alias
		base = f"/usr/local/bin"
		if not Files.exists(base):
			base = f"/usr/bin/"
		path = f"{base}/{alias}"
		sudo = Boolean("--sudo" in sys.argv).string(true="sudo ",false="")
		if venv != None and not Files.exists(venv):
			print(f"Creating virtual environment {venv}.")
			os.system(f"{self.vars.executable} -m venv {venv}")
			if requirements != None and Files.exists(requirements):
				print(f"Installing requirements {requirements}.")
				os.system(f"{venv}/bin/pip3 install -r {requirements} --user {self.vars.user}")
		if venv != None: venv = f'"{venv}"'
		file = f"""#!/usr/bin/env python3\nimport os, sys, platform\npackage="{executable}"\nvenv={venv}\nsys.argv.pop(0)\narguments = sys.argv\ns = ""\nfor i in arguments:\n	if s == "": \n		if " " in i: s = "'"+i+"'"\n		else: s = i\n	else: \n		if " " in i: s += " '"+i+"'"\n		else: s += " "+i\n"""+"""if venv != None: os.system(f"{venv}/bin/python3 "+package+" "+s)\nelif str(platform.system()).lower() in ["darwin"] and os.path.exists("/usr/bin/python3"): os.system("/usr/bin/python3 "+package+" "+s)\nelse:  os.system("python3 "+package+" "+s)\n#os.system("python3 "+package+" "+s)"""
		if ((cli.argument_present("--force") or cli.argument_present("--forced") or overwrite) and present) or (present or not Files.exists(path)) or (present and Files.exists(path) and Files.load(path) != file):
			if l_alias != None: alias = l_alias
			os.system(f"{sudo}touch {path}")
			os.system(f"{sudo}chmod +x {path}")
			os.system(f"{sudo}chown {self.vars.user}:{self.vars.group} {path}")
			try:
				File(path=f"{path}", data=file).save()
			except:
				print(f"Unable to create alias $ {alias}.")
				return None
			os.system(f"chmod +x {path}")
			if '--silent' not in sys.argv:
				print(f'Successfully created alias: {alias}.')
				print(f"Check out the docs for more info $: {alias} -h")
		if present:
			quit()

	# get source path.
	def source_path(self, path, back=1):
		source = gfp.clean(path=FilePath(path).base(back=back), remove_double_slash=True)
		if len(source) > 0 and source[len(source)-1] != "/": source += "/"
		return source

	# get log level.
	def log_level(self, default=0):
		return int(cli.get_argument("--log-level", required=False, default=env.get_integer("LOG_LEVEL", default=default)))

	# get current working directory.
	def pwd(self):
		return FilePath(utils.__execute_script__("pwd").replace("\n","")).clean()

	# insert package libraries.
	def insert(self, path):
		path = gfp.clean(path).replace(".","/")
		return sys.path.insert(1, path)

	# get the python site packages location
	def site_packages(self):
		path = FilePath(SOURCE_PATH).base()
		if FilePath(path).name() in ["site-packages", "dist-packages"]:
			return Directory(path)
		return Directory(sysconfig.get_paths()["purelib"])

	# install requirements.
	def install_requirements(self,
		# the requirements (#1).
		#	str instance: path to file.
		#	list instance: pip requirements in list
		requirements,
		# the silent option.
		silent=False,
		# the log level (Leave None to use self.options.log_level).
		log_level=None,
	):
		if log_level == None: log_level = self.options.log_level
		options = ""
		if isinstance(requirements, (str, String)):
			
			msg = f"Installing pip requirements [{requirements}]."
			if not os.path.exists(requirements):
				raise ValueError(f"Requirements file [{requirements}] does not exist.")
			if not os.path.isdir(requirements):
				options += "-r "
		elif isinstance(requirements, (list, Array)):
			msg = f"Installing {len(requirements)} pip requirement(s)."
			requirements = Array(requirements).string(joiner=" ")
		if log_level >= 0: print(msg)
		os.system(f"{self.vars.executable} -m pip install {options}{requirements} {Boolean(silent).string(true='2> /dev/null', false='')} --user {self.vars.user}")

	# interactive.
	def interactive(self, default=False):
		if cli.arguments_present(["--non-interactive"]): return False
		return env.get("INTERACTIVE", format=bool, default=default)

# initialized classes.
defaults = Defaults()
#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.files import *
from dev0s.classes.response import response as _response_
from dev0s.classes import code
from dev0s.classes.defaults.defaults import defaults

# the metrics object class.
class Metrics(object):
	def __init__(self):

		# docs.
		DOCS = {
			"module":"dev0s.system.metrics", 
			"initialized":True,
			"description":[], 
			"chapter": "Metrics", }

		#

	# get ram metrics.
	def ram(self, percentages=True):

		# vars.
		info = {}

		# get memory & swap.
		response = code.execute("""free -m | awk '{print $2"|"$3"|"$4"|"$5"|"$6}' """)
		if not response: return response
		output = response.output.split("\n")[1:]
		for index, type in [
			[0, "memory"],
			[1, "swap"],
		]:
			if type == "memory":
				total, used, free, shared, cache = output[index].split("|")
			else:
				total, used, free, _, _, = output[index].split("|")
				cache = 0.0
			if percentages:
				info[type] = {
					"free":round( (float(free) / float(total)) * 100, 2),
					"used":round( (float(used) / float(total)) * 100, 2),
					"cache":round( (float(cache) / float(total)) * 100, 2),
					"total":f"{total} MB",
				}
			else:
				info[type] = {
					"free":f"{free} MB",
					"used":f"{used} MB",
					"cache":f"{cache} MB",
					"total":f"{total} MB",
				}

		# handler.
		return _response_.success("Successfully retrieved the ram metrics.", {
			"info":info,
		})

	#





		



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.config import *
from dev0s.classes import code
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.response import response as _response_

# the wifi object class.
class WiFi(object):
	def __init__(self):

		# docs.
		DOCS = {
			"module":"dev0s.wifi", 
			"initialized":True,
			"description":[], 
			"chapter": "Network", }

		#

	# connect with a wifi network.
	def connect(self,
		# the network's ssid.
		ssid=None,
		# the network's password.
		password=None,
	):

		# linux.
		if defaults.vars.os in ["linux"]:
			path = f"/etc/netplan/{ssid}.yaml"
			Files.Save(path, f"""network:
		wifis:
	    	wlan0:
	        	dhcp4: true
	        	optional: true
	        	access-points:
	            	"{ssid}":
	            	    password: "{password}" """, sudo=True)
			response = code.execute("sudo netplan generate")
			if not response.success: return response
			response = code.execute("sudo netplan apply")
			if not response.success: return response
			return _response_.success(f"Successfully connected with {ssid}.")
		
		# invalid os.
		else:
			return _response_.error(f"Invalid operating system {defaults.vars.os}.")


# initialized classes.
wifi = WiFi()



#!/usr/bin/env python3
from dev0s.shortcuts import dev0s, Array, String
import sys
dev0s.response.log_timestamps = False
dev0s.response.log(f"&GREY&Comment:&PURPLE& {String(Array(sys.argv[1:]).string(joiner=' ')).capitalized_word()}&END&")



# imports
from dev0s.classes.config import *
from dev0s.classes.defaults.files import Date
from dev0s.classes.defaults.color import color, symbol
import getpass
_input_ = input

# log with safe replace.
def log(msg, back=0, log_timestamps=False):
	if log_timestamps: msg = f"{Date().seconds_timestamp} - {msg}"
	if back > 0:
		print(msg, end="\r")
	else:
		print(msg)
	#for _ in range(back):
	#	sys.stdout.write("\033[F") # cursor up one line
	sys.stdout.write("\033[K") # clear to the end of line

# input.
def input(message, yes_no=False, check=False, password=False, default=None):


	# docs.
	DOCS = {
		"module":"dev0s.console.input", 
		"description":[], 
		"chapter": "Console", }
		
	# attributes.
	message = color.fill(message)
	if len(message) > 0 and message[0].upper() != message[0]: message = String(message).capitalized_word()
	if yes_no:
		while True:
			if len(message) > 0 and message[len(message)-1] == " ": message = message[:-1]
			elif len(message) > 0 and message[len(message)-1] == ".": message = message[:-1]
			else: break
		value = _input_(message+" (y/n): ")
		if value in ["yes", 'y', "YES", "Y"]:
			return True
		else:
			return False
	else:
		while True:
			if len(message) > 0 and message[len(message)-1] == " ": message = message[:-1]
			elif len(message) > 0 and message[len(message)-1] == ":": message = message[:-1]
			else: break
		if password:
			value = getpass.getpass(message+": ")
		else:
			value = _input_(message+": ")
			if value in [""]: value = default
		if check:
			if password:
				if getpass.getpass("Enter the same passphrase: ") != password:
					log("Passphrases do not match.")
					return default
				else:
					return password
			else:
				if input(f"({value}) correct?", yes_no=True):
					return value
				else:
					return default
		else:
			return value

# log a divider.
def divider(dump=True):
	divider = ""
	for _ in range(int(os.get_terminal_size().columns)): divider += "-"
	if dump: print(divider)
	return divider

# the loader object class.
class Loader(threading.Thread):
	def __init__(self, 
		# the loader message (str) (#1).
		message:str, 
		# auto start the loader (bool).
		autostart=True, 
		# the log level (int).
		log_level=0, 
		# interactive (bool).
		interactive=True, 
		# log the timestamps (bool).
		log_timestamps=True,
	):

		# docs.
		DOCS = {
			"module":"Loader", 
			"initialized":False,
			"description":[], 
			"chapter": "Console", }
			
		# attributes.
		threading.Thread.__init__(self)
		self.message = self.__clean_message__(message)
		self.last_message = str(self.message)
		self.log_level = log_level
		self.interactive = interactive
		self.log_timestamps = log_timestamps
		if autostart and self.log_level >= 0: 
			if self.interactive:
				try:
					self.start()
				except KeyboardInterrupt as e:
					self.stop(success=False)
					raise KeyboardInterrupt(f"{e}")
			else:
				log(self.message+".")
	def run(self):
		if self.log_level >= 0: 
			self.running = True
			self.released = True
			while self.running:
				if not self.running: break
				elif not self.released:
					time.sleep(1)
				else:
					for i in ["|", "/", "-", "\\"]:
						if not self.released: break
						elif not self.running: break
						if self.message != self.last_message:
							#log(self.__empty_message__(length=len(f"{self.last_message} ...   ")), back=1)
							self.message = self.__clean_message__(self.message)
						log(f"{self.message} ... {i}", back=1, log_timestamps=self.log_timestamps) 
						self.last_message = self.message
						if not self.released: break
						elif not self.running: break
						time.sleep(0.2)
		self.running = "stopped"
		sys.exit(0)
	def stop(self, message=None, success=True, response=None, quiet=False):
		if self.log_level >= 0:
			if response == None:
				if message == None: message = self.message
			else:
				if response["error"] == None:
					message = response["message"]
				else:
					success = False
					message = "Error: "+response["error"]
			if self.interactive:
				self.running = False
				for i in range(120):
					if self.running == "stopped": break
					time.sleep(0.2)
				if self.running != "stopped": raise ValueError(f"Unable to stop loader [{self.message}].")
			if not quiet:
				if self.interactive:
					#log(self.__empty_message__(length=len(f"{self.last_message} ...   ")), back=1)
					if success:
						log(f"{message} ... done", log_timestamps=self.log_timestamps)
					else:
						log(f"{message} ... {color.red}failed{color.end}", log_timestamps=self.log_timestamps)
				else:
					if success:
						log(f"{message}. done", log_timestamps=self.log_timestamps)
					else:
						log(f"{message}. {color.red}failed{color.end}", log_timestamps=self.log_timestamps)
	def mark(self, new_message=None, old_message=None, success=True, response=None):
		if self.log_level >= 0: 
			if response != None:
				if response["error"] == None:
					success = True
				else:
					success = False
			if old_message == None: old_message = self.message
			if self.interactive:
				#log(self.__empty_message__(length=len(f"{self.last_message} ...   ")), back=1)
				if success:
					log(f"{old_message} ... done", log_timestamps=self.log_timestamps)
				else:
					log(f"{old_message} ... {color.red}failed{color.end}", log_timestamps=self.log_timestamps)
			else:
				
				if success:
					log(f"{old_message}. done", log_timestamps=self.log_timestamps)
				else:
					log(f"{old_message}. {color.red}failed{color.end}", log_timestamps=self.log_timestamps)
			if new_message != None: self.message = new_message
	def hold(self):
		if self.log_level >= 0: 
			self.released = False
			time.sleep(0.33)
	def release(self):
		if self.log_level >= 0: 
			self.released = True
			time.sleep(0.33)
	# system functions.
	def __clean_message__(self, message):
		message = color.fill(str(message))
		if message[-len(" ..."):] == " ...": message = message[:-4]
		if message[-len("."):] == ".": message = message[:-1]
		if message[0].upper() != message[0]: message = message[1:]+message[0].upper()+message[1:]
		return message
	def __empty_message__(self, length=len("hello world")):
		s = ""
		for i in range(length): s += " "
		return s

# the loader object class.
class ProgressLoader(object):
	def __init__(self, 
		# the message (str) (#1).
		message:str, 
		# the start index (pass shared multiprocessing variable to use with mp) (int).
		index=0, 
		# the max index (int).
		max=10, 
		# calculate estimated finish time.
		calc_finish=False,
		# auto start.
		autostart=True,
		# log the timestamps.
		log_timestamps=True,
		# log the duration when finished.
		log_duration=False,
		# the memory (dict) (for multiprocessing) (leave None to ignore).
		# uses the key (loader_index) as the loader's index value.
		# increment the loader index from within the processes and call loader.wait() to await till the loader is finished.
		memory=None,
		# the total amount of processes (int) (required for multiprocessing)
		processes=None,
		# the active log level (int).
		log_level=0,
	):

		# docs.
		DOCS = {
			"module":"ProgressLoader", 
			"initialized":False,
			"description":[], 
			"chapter": "Console", }
			
		# attributes.
		self.message = message
		if self.message[-len(" ..."):] == " ...": self.message = self.message[:-4]
		if self.message[-len("."):] == ".": self.message = self.message[:-1]
		if self.message[0].upper() != self.message[0]: self.message = self.message[1:]+self.message[0].upper()+self.message[1:]
		self.message = color.fill(self.message)
		self.index = index
		self.max = max
		self.calc_finish = calc_finish
		self.start_stamp = None
		self.progress = None
		self.last_message = None
		self.log_level = log_level
		self.log_timestamps = log_timestamps
		self.log_duration = log_duration
		self.memory = memory
		self.processes = processes
		if self.memory not in [None, False]:
			if self.processes in [None, False] or self.processes < 0: raise ValueError("Multiprocessing is enabled by defining the memory parameter while likely forgot to define the number of processes parameter (processes:int).")
			if self.index != 0: raise ValueError("A multiprocessing progress loader should always start with index 0.")
			if "loader_index" not in list(self.memory.keys()): 
				raise ValueError("The multiprocessing memory dict must include the key (loader_index) with a multiprocessing dict as value.")
			for id in range(self.processes): self.memory["loader_index"][id] = 0
			try: self.memory["loader_index"][0]
			except KeyError:
				raise ValueError("The multiprocessing memory dict must include the key (loader_index) with a multiprocessing dict as value.")
		if autostart: self.next(increment=0)

	# next index.
	def next(self, increment=1, decimals=2):
		if self.calc_finish and self.start_stamp == None:
			self.start_stamp = Date()
		if self.memory in [None, False]:
			self.index += increment
			try:
				p = round((self.index / self.max) * 100, decimals)
			except ZeroDivisionError:
				p = 0.0
		else:
			if increment > 0: raise ValueError("Unable to increment the index with loader.next() when multiprocessing is enabled.")
			t = 0
			for id in range(self.processes): t += self.memory["loader_index"][id]
			try:
				p = round((t / self.max) * 100, decimals)
			except ZeroDivisionError:
				p = 0.0
		if p != self.progress:
			self.progress = p
			if self.log_timestamps:
				self.last_message = f"{Date().seconds_timestamp} - {self.message} ... {self.progress}%"
			else:
				self.last_message = f"{self.message} ... {self.progress}%"
			if self.calc_finish and self.progress > 0:
				diff = (Date() - self.start_stamp).to_seconds()
				duration = Date().normalize_seconds( (diff / (self.progress/100)) * (1.0 - (self.progress/100)) )
				self.last_message += f" ({duration})"
			if self.log_level >= 0:
				log(self.last_message, back=1)
		#

	# stop the loader.
	def stop(self, message=None, success=True, response=None):
		if self.log_level >= 0:
			if response == None:
				if message == None: message = self.message
			else:
				if response["error"] == None:
					message = response["message"]
				else:
					success = False
					message = "Error: "+response["error"]
			if self.last_message != None:
				log(self.__empty_message__(length=len(f"{self.last_message}")), back=1)
			s = ""
			if success:
				s = f"{message} ... done"
				if self.log_duration:
					duration = Date().normalize_seconds(Date().to_seconds() - self.start_stamp.to_seconds())
					s += f" (duration: {duration})"
			else:
				s = f"{message} ... {color.red}failed{color.end}"
			log(s, log_timestamps=self.log_timestamps)
		#

	# wait till the loader is finished the progress (for multiprocessing).
	# make sure self.max is the actual max otherwise it will run forever if timeout is None.
	def wait(self,
		# the sleep time (int).
		sleeptime=15,
		# timeout in seconds (leave None for no timeout) (int).
		timeout=None,
	):
		seconds = 0
		while True:
			self.next(increment=0)
			if self.progress >= 100: return True
			if timeout != None and seconds > timeout: break
			seconds += sleeptime
			time.sleep(sleeptime)
		raise ValueError("Timeout expired.")

	# create empty message.
	def __empty_message__(self, length=len("hello world")):
		s = ""
		for i in range(length): s += " "
		return s
		#

	#

#

#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.files import *
from dev0s.classes.response import response as _response_
from dev0s.classes import code
from dev0s.classes.defaults.defaults import defaults

# the cleanup object class.
class CleanUp(object):
	def __init__(self):

		# docs.
		DOCS = {
			"module":"dev0s.system.cleanup", 
			"initialized":True,
			"description":[], 
			"chapter": "CleanUp", }

		#

	# clean the ram memory.
	def ram(self):

		# linux.
		if defaults.vars.os in ["linux"]:
			path = f"/tmp/ram_cleanup.sh"
			Files.save(path, '''

				#!/bin/bash
				## Bash Script to clear cached memory on (Ubuntu/Debian) Linux
				## By Philipp Klaus
				## see <http://blog.philippklaus.de/2011/02/clear-cached-memory-on-ubuntu/>
				
				if [ "$(whoami)" != "root" ]
				then
				  echo "You have to run this script as Superuser!"
				  exit 1
				fi

				# Get Memory Information
				freemem_before=$(cat /proc/meminfo | grep MemFree | tr -s ' ' | cut -d ' ' -f2) && freemem_before=$(echo "$freemem_before/1024.0" | bc)
				cachedmem_before=$(cat /proc/meminfo | grep "^Cached" | tr -s ' ' | cut -d ' ' -f2) && cachedmem_before=$(echo "$cachedmem_before/1024.0" | bc)

				# Output Information
				#echo -e "This script will clear cached memory and free up your ram.\\n\\nAt the moment you have $cachedmem_before MiB cached and $freemem_before MiB free memory."

				# Test sync
				if [ "$?" != "0" ]
				then
				  echo "Something went wrong, It's impossible to sync the filesystem."
				  exit 1
				fi

				# Clear Filesystem Buffer using "sync" and Clear Caches
				sync && echo 3 > /proc/sys/vm/drop_caches

				freemem_after=$(cat /proc/meminfo | grep MemFree | tr -s ' ' | cut -d ' ' -f2) && freemem_after=$(echo "$freemem_after/1024.0" | bc)

				# Output Summadry
				#echo -e "This freed $(echo "$freemem_after - $freemem_before" | bc) MiB, so now you have $freemem_after MiB of free RAM."

			''')
			Files.chmod(path, "+x")
			output = code.execute(f"sudo {path}")
			#Files.delete(path, forced=True)
			if not output.success: return output

		# macos.
		#elif defaults.vars.os in ["linux"]:

		# invalid os.
		else: raise ValueError(f"Unsupported operating system: {defaults.vars.os}.")
		

		# handler.
		return _response_.success("Successfully cleared the RAM memory.")

	#

	# clean the swap memory.
	def swap(self):

		# linux.
		if defaults.vars.os in ["linux"]:
			output = code.execute(f"""
				sudo swapoff -a
				sudo swapon -a
			""")
			if not output.success: return output

		# macos.
		#elif defaults.vars.os in ["linux"]:

		# invalid os.
		else: raise ValueError(f"Unsupported operating system: {defaults.vars.os}.")
		

		# handler.
		return _response_.success("Successfully cleared the swap memory.")

	#





		



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.files import *
from dev0s.classes.manager import installation
from dev0s.classes import console

# the manager object class.
class Manager(object):
	def __init__(self):	
		
		# objects.
		self.installation = installation.Installation()
		self.lib = f"{SOURCE_PATH}/lib/"
		self.user_lib = "/usr/local/lib/"
		self.user_bin = "/usr/local/bin/"
		for path in [self.user_bin, self.user_lib]:
			if not Files.exists(path): os.system(f"sudo mkdir -p {path} && sudo chown {OWNER}:{GROUP} {path} && sudo chmod 770 {path}")
		self.check()

		#
	def check(self):

		# only if installed.
		if self.installation.__installed__():

			# check python lib.
			gfp = FilePath("")
			for lib in Files.Directory(f"{self.lib}/python/").paths(recursive=False, files_only=True, banned_names=[".DS_Store", "example"]):
				current = File(path=gfp.clean(path=lib, remove_last_slash=True))
				installed = File(path=gfp.clean(path=f"{self.user_bin}/{gfp.name(path=lib)}", remove_last_slash=True))
				install = False
				if not installed.file_path.exists(): install = True
				elif installed.load() != current.load(): install = True
				if install: 
					loader = console.Loader(f"Installing library {installed.file_path.name()}")
					os.system(f"sudo rm -fr {installed.file_path.path} && sudo cp -r {current.file_path.path} {installed.file_path.path} && sudo chmod +x {installed.file_path.path} && sudo chown {OWNER}:{GROUP} {installed.file_path.path}")
					if not installed.file_path.exists(): loader.stop(success=False)
					else: loader.stop()

# initialized objects.
manager = Manager()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.exceptions import Exceptions
from dev0s.classes import utils
from dev0s.classes import console
from dev0s.classes.defaults.files import *
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.response import response as _response_

# include "execute" functions & classes.
from dev0s.classes.code.execute import processes, kill, execute, Spawn, OutputObject

# askpass still to create.
def askpass():
	if defaults.vars.os in ["linux"]:
		askpass = f"""echo '#!/bin/bash' > /tmp/askpass.sh && echo 'zenity --password --title="{ALIAS} requires root permissions."' >> /tmp/askpass.sh && export SUDO_ASKPASS="/tmp/askpass.sh" """
		output = dev0s.utils.__execute_script__(f"cd ~ && rm -fr /tmp/{ALIAS}/ && git clone -q https://github.com/vandenberghinc/{ALIAS} /tmp/{ALIAS}/ && chmod +x /tmp/{ALIAS}/requirements/installation.sh && {askpass} && sudo -A -u {user} bash /tmp/{ALIAS}/requirements/installation.sh")
	else:
		askpass = f"""/usr/bin/osascript -e 'do shell script "/tmp/{ALIAS}/requirements/installation.sh" with administrator privileges'"""

# the version object class.
class Version(object):
	def __init__(self, 
		# the version value (#1).
		value="1.0.00",
		# the file path (optional) (str, FilePath).
		path=None,
		# the default value (opional) (str)
		default=None,
		# load the specified path data on initialization (bool).
		load=False,
	):

		# docs.
		DOCS = {
			"module":"Version", 
			"initialized":False,
			"description":[], 
			"chapter": "Code", }

		# defaults.
		#self.__class__.__name__ = "Version"

		# check self instance.
		if isinstance(value, Version):
			self.value = value.value

		# init.
		self.value = str(value).replace(" ","").replace("\n","").replace("\r","")
		self.int = int(str(self.value).replace(".",""))

		# path.
		if path == False: self.file_path = self.fp = None # used in local memory (not fysical)
		else: self.file_path = self.fp = Formats.FilePath(path)
		if default != None and not Files.exists(self.file_path.path): self.save(array=default)
		if load: self.load()

		#
	def save(self, value=None, path=None, sudo=False):
		if value == None: value = self.value
		if path == None: path = self.file_path.path
		utils.__check_memory_only__(path)
		self.assign(string)
		return Files.save(path, str(string), format="str", sudo=sudo)
	def load(self, default=None, sudo=False):
		utils.__check_memory_only__(self.file_path.path)
		if not os.path.exists(self.file_path.path) and default != None: 
			self.save(default, sudo=sudo)
		self.assign(Files.load(self.file_path.path, format="str", sudo=sudo))
		return self.value
	def increase(self, value=None, count=1):
		if value == None: value = self.value

		# version 2.
		path = "/tmp/increase_version"
		Files.save(path, """version=''$1'' && echo $version | awk -F. -v OFS=. 'NF==1{print ++$NF}; NF>1{if(length($NF+1)>length($NF))$(NF-1)++; $NF=sprintf("%0*d", length($NF), ($NF+1)%(10^length($NF))); print}'""")
		for i in range(int(count)):
			value = subprocess.check_output([f"bash", path, str(value)]).decode().replace("\n","")
		self.assign(value=value)
		return value

		# version 1.
		#
		old_version = value
		base, _base_= [], old_version.split(".")
		increase = True
		for i in _base_:
			base.append(int(i))
		count = len(base)-1
		for i in range(len(base)):
			if increase:
				if base[count] >= 9:
					if count > 0:
						base[count-1] += 1
						base[count] = 0
						increase = False
					else:
						base[count] += 1
						break
				else:
					base[count] += 1
					break
			else:
				if count > 0 and int(base[count]) >= 10:
					base[count-1] += 1
					base[count] = 0
					increase = False
				elif count == 0: break
			count -= 1
		version = ""
		for i in base:
			if version == "": version = str(i)
			else: version += "."+str(i) 
		return version
		#
	# int format.
	def __index__(self):
		return self.int
	# support "+, -, +=, -=" .
	def __add__(self, count):
		if isinstance(count, (int, float)):
			a=1
		elif isinstance(count, Integer):
			count = int(count.value)
		else:
			raise Exceptions.FormatError(f"Can not add object {self.__class__} & {count.__class__}, a version should be incremented with a count (example: 1).")
		value = self.increase(value=self.value, count=count)
		return Version(value)
	def __concat__(self, count):
		if isinstance(vercount, (int, float)):
			a=1
		elif isinstance(count, Integer):
			count = int(count.value)
		else:
			raise Exceptions.FormatError(f"Can not concat object {self.__class__} & {count.__class__}, a version should be incremented with a count (example: 1).")
		value = self.increase(value=self.value, count=count)
		return Version(value)
	def __pos__(self, count):
		if isinstance(count, (int, float)):
			a=1
		elif isinstance(count, Integer):
			count = int(count.value)
		else:
			raise Exceptions.FormatError(f"Can not pos object {self.__class__} & {count.__class__}, a version should be incremented with a count (example: 1).")
		value = self.increase(value=self.value, count=count)
		return Version(value)
	def __iadd__(self, count):
		if isinstance(count, (int, float)):
			a=1
		elif isinstance(count, Integer):
			count = int(count.value)
		else:
			raise Exceptions.FormatError(f"Can not add object {self.__class__} & {count.__class__}, a version should be incremented with a count (example: 1).")
		self.assign(self.increase(value=self.value, count=count))
		return self
	def __sub__(self, count):
		raise Exceptions.InvalidUsage("A version can not be decremented.")
	def __isub__(self, count):
		raise Exceptions.InvalidUsage("A version can not be decremented.")

	# support default iteration.
	def __iter__(self):
		return iter(self.value.split("."))
	
	# support '>=' & '>' & '<=' & '<' operator.
	def __gt__(self, version):
		if not isinstance(version, self.__class__):
			raise Exceptions.FormatError(f"Can not compare object {self.__class__} & {version.__class__}.")
		else:
			version = version.int
		return self.int > version
	def __ge__(self, version):
		if not isinstance(version, self.__class__):
			raise Exceptions.FormatError(f"Can not compare object {self.__class__} & {version.__class__}.")
		else:
			version = version.int
		return self.int >= version
	def __lt__(self, version):
		if not isinstance(version, self.__class__):
			raise Exceptions.FormatError(f"Can not compare object {self.__class__} & {version.__class__}.")
		else:
			version = version.int
		return self.int < version
	def __le__(self, version):
		if not isinstance(version, self.__class__):
			raise Exceptions.FormatError(f"Can not compare object {self.__class__} & {version.__class__}.")
		else:
			version = version.int
		return self.int <= version
	
	# support '==' & '!=' operator.
	def __eq__(self, version):
		if isinstance(version, None.__class__):
			return False
		elif isinstance(version, self.__class__):
			return self.int == version.int
		else:
			return str(self.int) == str(version)
	def __ne__(self, version):
		if isinstance(version, None.__class__):
			return True
		elif isinstance(version, self.__class__):
			return self.int != version.int
		else:
			return str(self.int) != str(version)
	
	# support 'in' operator.
	def __contains__(self, string):
		if isinstance(string, (list, Files.Array)):
			for i in string:
				if str(i) in str(self.value):
					return True
			return False
		else:
			return str(string) in str(self.value)
		#
	# representation.
	def __repr__(self):
		return str(self)
		#
	# str representation.
	def __str__(self):
		return str(self.value)
	# int representation.
	def __int__(self):
		return self.int
	# float representation.
	def __float__(self):
		return float(self.int)
	# content count.
	def __len__(self):
		return len(str(self.value))
	# object id.
	def __id__(self):
		return f"({self.instance()}:{str(self)})"
	# object instance.
	def instance(self):
		return "Version"
		#
	# support self assignment.
	def assign(self, value):
		if isinstance(value, (str, String)):
			a=1
		elif isinstance(value, self.__class__):
			value = value.value
		elif not isinstance(value, self.__class__):
			raise Exceptions.FormatError(f"Can not assign object {self.__class__} & {value.__class__}.")
		self.value = str(value)
		self.int = int(str(self.value).replace(".",""))
		return self
	# return raw data.
	def raw(self):
		return self.value
	#

# the script object class.
class Script(Files.File):
	def __init__(self, 
		# the script data (str) (param #1).
		data=None, 
		# the path (str) (param #2).
		path=None, 
		# the script executable.
		executable="sh",
		# load the data.
		load=False, 
		# the default data (create if path does not exist).
		default=None,
	):
		
		# docs.
		DOCS = {
			"module":"Script", 
			"initialized":False,
			"description":[], 
			"chapter": "Code", }

		# check self instance.
		if isinstance(data, (Files.File, Script)):
			data = data.data
		if data == None: data = ""
		Files.File.__init__(self,
			path=path,
			data=data,
			load=load,
			default=default,
		)
		self.executable = executable

		#

	# execute.
	def execute(self, 
		# the script executable (leave None to use deafult self.executable) (#1).
		executable=None,
		# the script path (str) (leave None to use default self.path) (#2).
		path=None,
	):
		if data == None: data = self.data
		if executable == None: executable = self.executable
		return execute(path=data, executable=executable)

		#

	# object instance.
	def instance(self):
		return "Script"
	@property
	def __name__(self):
		return self.instance()
		#
	# support self assignment.
	def assign(self, string):
		if isinstance(data, (int, float)):
			a=1
		elif isinstance(data, (String)):
			data = data.string
		elif isinstance(data, self.__class__):
			data = data.data
		elif not isinstance(data, self.__class__):
			raise Exceptions.FormatError(f"Can not assign object {self.__class__} & {data.__class__}.")
		self.data = str(data)
		return self
	#

# the python object class.
class Python(Files.File):
	def __init__(self, 
		# the script data (str) (param #1).
		data=None, 
		# the path (str) (param #2).
		path=None, 
		# the script executable.
		executable="python3",
		# load the data.
		load=False, 
		# the default data (create if path does not exist).
		default=None,
	):

		# docs.
		DOCS = {
			"module":"Python", 
			"initialized":False,
			"description":[], 
			"chapter": "Code", }

		# check self instance.
		if isinstance(data, (Files.File, Script, Python)):
			data = data.data
		if data == None: data = ""
		self.directory = False
		if path != None and gfp.exists(path=path) and gfp.directory(path=path):
			self.directory = True
			load = False
		Files.File.__init__(self,
			path=path,
			data=data,
			load=load,
			default=default,
		)
		self.executable = executable
		
		#
		#
	# execute.
	def execute(self, 
		# the script executable (leave None to use defult self.executable) (#1).
		executable=None,
		# the script path (str) (leave None to use default self.path) (#2).
		path=None,
	):
		if data == None: data = self.data
		if executable == None: executable = self.executable
		return execute(path=data, executable=executable)

		#
		#
	# slice classes.
	def slice_classes(self,
		# the data (str).
		data=None, 
		# the wanted indent of the init line (optional).
		indent=None,
		# skip unparsable functions & classes.
		skip_errors=False,
		# the banned classes.
		banned_classes=[],
		# the banned functions.
		banned_functions=[],
		# the module name to use before the classes (optional) (defaults.Traceback always overwrites).
		module=None,
		# get inside classes.
		inside_classes=True,
		# get inside functions (recursive is always False).
		inside_functions=True,
		# include system functions (aka: def __somefunc__).
		system_functions=False,
		# only for self recursive use.
		keep_indent=False,
	):

		# checks.
		if data == None: data = self.data
		if data == None: raise Exceptions.InvalidUsage("Define parameter: [data].")

		# normalize.
		data = data.replace("	","    ")
		before = "\n"
		if len(data) > 0 and data[0] != before: data = before+str(data)
		while True:
			if "	" in data: data = data.replace("	","    ")
			elif ") :" in data: data = data.replace(") :","):")
			elif ")  :" in data: data = data.replace(")  :","):")
			elif ")   :" in data: data = data.replace(")   :","):")
			elif ")    :" in data: data = data.replace(")    :","):")
			elif ")     :" in data: data = data.replace(")     :","):")
			elif ")      :" in data: data = data.replace(")      :","):")
			elif ")       :" in data: data = data.replace(")       :","):")
			elif "): \n" in data: data = data.replace("): \n","):\n")
			elif "):  \n" in data: data = data.replace("):  \n","\n")
			elif "):   \n" in data: data = data.replace("):   \n","):\n")
			elif "):    \n" in data: data = data.replace("):    \n","):\n")
			elif "):     \n" in data: data = data.replace("):     \n","):\n")
			elif "):      \n" in data: data = data.replace("):      \n","):\n")
			elif "):       \n" in data: data = data.replace("):       \n","):\n")
			elif "):        \n" in data: data = data.replace("):        \n","):\n")
			#elif "\n " in data: data = data.replace("\n ", "\n")
			elif " \n" in data: data = data.replace(" \n", "\n")
			elif "DOCS =  " in data: data = data.replace("DOCS =  ", "DOCS = ")
			elif "DOCS  = " in data: data = data.replace("DOCS  = ", "DOCS = ")
			elif "DOCS   = " in data: data = data.replace("DOCS   = ", "DOCS = ")
			elif "DOCS    = " in data: data = data.replace("DOCS    = ", "DOCS = ")
			elif "DOCS     = " in data: data = data.replace("DOCS     = ", "DOCS = ")
			else: break

		# vars.
		classes = []
		to_slice = []
		for i in range(100):
			l = "\n"
			for _ in range(i): l += " "
			l += "class "
			to_slice.append(l)
		slices = String(data).split(to_slice)
		if len(slices) == 1:
			if format == dict:
				return {}
			else:
				return ""
		elif len(slices) > 1:
			slices = slices[1:]
		slice_count = 0

		# slice.
		for class_ in slices:
			if class_ not in ["", " "]:
				try: previous = slices[slice_count-1]
				except IndexError: previous = None
				if previous == None:
					init_line_indent = 0
				else:
					init_line_indent = 0
					l = len(previous)
					for _ in range(len(previous)):
						l -= 1
						if previous[l] != " ": break
						init_line_indent += 1
					if init_line_indent > 0: init_line_indent += 1 # beause of the space after [class ].
				if indent == None or str(init_line_indent) == str(int(indent)):
					class_ = String("class "+class_)
					class_line = class_.split("\n")[0]
					class_name = class_.split("(")[0].replace("class ","")
					if class_name not in banned_classes:
						if module != None:
							if module[len(module)-1] == ".": module = module[:-1]
							class_name = f"{module}.{class_name}"
						if class_name not in banned_classes:
							if "\n" not in class_:
								class_indent = init_line_indent+1
							else:
								class_indent = class_.line_indent(line=class_.split("\n")[1])
							class_code = class_.slice_indent(indent=class_indent, depth=1, remove_indent=True)
							raw_class_code = class_.slice_indent(indent=class_indent, depth=1, remove_indent=False)
							class_type = String(class_).slice_tuple(depth=1)[1:-1].replace(" ","").replace(",",", ")
							class_parameters = String(class_code).slice_tuple(depth=1)
							#print("class_name:".upper(),class_name)
							#print("class_type:".upper(),class_type)
							#print("class_parameters:".upper(),class_parameters)
							raw_name = str(class_name)
							if "." in raw_name: raw_name = raw_name.split('.')[len(raw_name.split('.'))-1]
							l_classes, functions = [], []
							recursive_clases = 0
							if inside_classes:
								while True:
									if ("\nclass " in class_code or " class " in class_code or (len(class_code) >= len("class ") and class_code[:len("class ")] == "class ")) and class_code != data and len(class_code) > 0:
										# get the depth of the class inside the class.
										inside_class_code_indent = String(class_code).line_indent(line=class_code.split("\n")[1])
										inside_class_code_depth = 0
										for line in class_code.split("\n"):
											if str(String(line).line_indent(line=line)) == str(0):
												inside_class_code_depth += 1
											if (len(line) >= len("class ") and line[:len("class ")] == "class ") or "\nclass " in line or " class " in line: 
												break
											#if (len(line) >= len("def ") and line[:len("def ")] == "def ") or "\ndef " in line or " def " in line: 
											#	inside_class_code_depth += 1
											#elif (len(line) >= len("class ") and line[:len("class ")] == "class ") or "\nclass " in line or " class " in line: 
											#	inside_class_code_depth += 1
											#	break
										inside_class_line = "class "+(class_.split(" class ")[1].split("\n")[0])
										#print("inside_class_line:".upper(),inside_class_line)
										#print("inside_class_code_indent:".upper(),inside_class_code_indent)
										#print("inside_class_code_depth:".upper(),inside_class_code_depth)
										#print(f"CLASS_CODE: [\n{class_code}]")
										inside_class_code = String(class_code).slice_indent(indent=inside_class_code_indent, depth=inside_class_code_depth, remove_indent=False)
										clean_inside_class_code = String(class_code).slice_indent(indent=inside_class_code_indent, depth=inside_class_code_depth, remove_indent=True)
										a = inside_class_line.split("(")
										a.pop(0)
										inside_class_type = String("("+str(Array(a).string(joiner=" "))+str(inside_class_code)).slice_tuple(depth=1).replace(" ","").replace(",",", ")
										#print(f"BEFORE CLASS_CODE: [\n{class_code}]")
										#print(f"INSIDE_CLASS_CODE: [\n{inside_class_code}]")
										old = str(class_code)
										full_inside_class = f"{inside_class_line.split('(')[0]}{inside_class_type}:\n{inside_class_code}"
										class_code = class_code.replace("):  \n","):\n").replace("):   \n","):\n").replace("):    \n","):\n").replace("):     \n","):\n").replace("):	\n","):\n").replace(full_inside_class, "")
										if old == class_code:
											if skip_errors:
												break
											else:
												raise ValueError(f"Replaced class code is the same as old class code: [{full_inside_class}]")
										#print(f"AFTER CLASS_CODE: [\n{class_code}]")
										#quit()
										l_module = module + "." + class_name
										classes += self.slice_classes(data=full_inside_class, keep_indent=keep_indent, inside_classes=inside_classes, inside_functions=True, system_functions=system_functions, banned_functions=banned_functions, banned_classes=banned_classes, module=l_module)
										#print(f"{inside_class_line}\n{clean_inside_class_code}")
										#quit()
										if ("\nclass " in class_code or " class " in class_code or (len(class_code) >= len("class ") and class_code[:len("class ")] == "class ")) and class_code != data and len(class_code) > 0:
											recursive_clases += 1
										else:
											break
									else:
										break
									#print(f"detected new class [{class_code}] from code [{class_}] ")
							if inside_functions:
								if ("\ndef " in class_code or " def " in class_code or (len(class_code) >= len("def ") and class_code[:len("def ")] == "def ")) and class_code != data and len(class_code) > 0:
									functions = []
									#l_module = module + "." + String(class_name).variable_format()
									l_module = String(raw_name).variable_format()
									for i in self.slice_functions(inside_classes=False, inside_functions=False, data=class_code, indent=0, keep_indent=False, system_functions=system_functions, banned_functions=banned_functions, banned_classes=banned_classes, module=l_module):
										skip = False
										old = str(class_code)
										class_code = class_code.replace(str(i["raw"]), "")
										#class_code = class_code.replace(i["raw_full_code"], "")
										#class_code = class_code.replace(i["init_line"].split("(")[0]+i["parameters"], "")
										#class_code = class_code.replace(i["raw_code"], "")
										if class_code == old:
											if skip_errors:
												skip = True
											else:
												raise ValueError(f"Replaced class code is the same as old class code, includes: {i['raw_full_code'] in class_code}, replacement: [{i['raw_full_code']}], code: [{class_code}].")
										if i["classes"] != {}:
											classes += i["classes"]
										del i["classes"]
										if not skip:
											functions.append(i)
							classes.append({
								"name":str(class_name),
								"raw_name":str(raw_name),
								"raw":class_,
								"code":str(class_code),
								"raw_code":str(raw_class_code),
								"indent":int(class_indent),
								"init_line":str(class_line),
								"parameters":str(class_parameters),
								"class_type":str(class_type),
								"functions":functions,
								"type":"class",
							})
			slice_count += 1

		# fill DOCS customization.
		_classes_ = []
		for info in classes:
			initialized, module, description, chapter = False, None, [], None
			if len(info["functions"]) > 0:
				for func in info["functions"]:
					if "__init__" in func["name"]:
						test = "DOCS = {"
						if test in func["code"]:
							customization = ast.literal_eval(str(String(func["code"].split(test[:-1])[1]).slice_dict(depth=1)))
							if "initialized" in customization:
								info["initialized"] = customization["initialized"]
							if "module" in customization:
								info["module"] = customization["module"]
							if "notes" in customization:
								info["description"] = customization["description"]
							if "chapter" in customization:
								info["chapter"] = customization["chapter"]
							#if "return" in customization:
							#	info["return"] = customization["return"]
							break
			for key,value in {
				"initialized":False,
				"module":None,
				"description":[],
				"chapter":None,
				#"return":"_",
			}.items():
				try:info[key]
				except: info[key] = value
			_classes_.append(info)
		classes = _classes_

		# handler.
		return classes

		#

	# slice functions.
	def slice_functions(self,
		# the data (str).
		data=None, 
		# the wanted indent of the init line (optional).
		indent=None,
		# for self recursive use.
		keep_indent=False,
		# get inside classes.
		inside_classes=True,
		# get inside functions (recursive is always False).
		inside_functions=False,
		# include system functions (aka: def __somefunc__).
		system_functions=False,
		# skip unparsable functions & classes.
		skip_errors=False,
		# the banned classes.
		banned_classes=[],
		# the banned functions.
		banned_functions=[],
		# the module name to use before the 1st indent functions (optional).
		module=None,
	):	
		# func.
		def parse_variable_format(variable, code):
			def parse_value(value):
				if len(value) > 0 and value[0] == "[":
					return "list"
				elif len(value) > 0 and value[0] == "{":
					return "dict"
				elif len(value) > 0 and value[0] in ["'", '"']:
					return "str"
				else:
					for i, return_ in [
						["None", "_"],
						["str", "str"],
						["int", "int"],
						["float", "float"],
						["list", "list"],
						["dict", "dict"],
						["FilePath", "FilePath"],
						["String", "String"],
						["Integer", "Integer"],
						["Boolean", "Boolean"],
						["Array", "Array"],
						["Dictionary", "Dictionary"],
						["response.success", "response"],
						["response.error", "response"],
					]:
						if len(value) >= len(i) and value[:len(i)] == i:
							return return_
					try: 
						int(value.split("\n")[0])
						if "." in value.split("\n")[0]:
							try:
								float(value.split("\n")[0])
								return "float"
							except: return "int"
						else:
							return "int"
					except: a=1
					return "parameter:"+value.split("\n")[0]
				#

			code = code.replace(" =","=").replace("= ","=").replace(" =","=").replace("= ","=").replace(" =","=").replace("= ","=").replace(" =","=").replace("= ","=").replace(" ,",",").replace(", ",",").replace(" ,",",").replace(", ",",").replace(" ,",",").replace(", ",",").replace(" ,",",").replace(", ",",")
			return_ = None
			# version 2.
			lines = code.split("\n")
			l = len(lines)
			for _ in range(len(lines)):
				l -= 1
				line = lines[l]
				if line.count("=") == 1 and variable in line:
					before,after = line.split("=")
					before = before[String().line_indent(line=before):]
					local = None
					if "," in before:
						if variable in before: # must include in before = because it assigns.
							c = None
							for i in before.split(","):
								if c == None: c = 0
								if i == variable:
									break
								c += 1
							if c == None:
								raise ValueError(f"Unable to find the tuple place of variable [{variable}] in tuple [{before}].")
							value = after.split(",")[c]
							if value != variable:
								local = parse_value(value)
								if len(local) >= len("parameter:") and local[:len("parameter:")] == "parameter:":
									new_var = local[len("parameter:"):]
									if new_var != variable:
										return parse_variable_format(new_var, code)
									else:
										local = None
					else:
						if after != variable: # aka $var = ... / not $other_var = $var
							local = parse_value(after)
							if len(local) >= len("parameter:") and local[:len("parameter:")] == "parameter:":
								new_var = local[len("parameter:"):]
								if new_var != variable:
									return parse_variable_format(new_var, code)
								else:
									local = None
					if local != None:
						return_ = local
						break
			return return_
			#
		def func_allowed(name, strip=False):
			if strip and "." in name: name = name.split(".")[len(name.split("."))-1]
			if name in banned_functions: return False
			if not system_functions and "__init__" not in name and len(name) >= 4 and name[:2] == "__" and name[-2:] == "__": return False
			return True
			#

		# vars.
		functions = []
		if data == None: data = self.data
		if data == None: raise Exceptions.InvalidUsage("Define parameter: [data].")
		
		# normalize.
		data = data.replace("	","    ")
		before = "\n"
		if len(data) > 0 and data[0] != before: data = before+str(data)
		while True:
			if "	" in data: data = data.replace("	","    ")
			elif ") :" in data: data = data.replace(") :","):")
			elif ")  :" in data: data = data.replace(")  :","):")
			elif ")   :" in data: data = data.replace(")   :","):")
			elif ")    :" in data: data = data.replace(")    :","):")
			elif ")     :" in data: data = data.replace(")     :","):")
			elif ")      :" in data: data = data.replace(")      :","):")
			elif ")       :" in data: data = data.replace(")       :","):")
			elif "): \n" in data: data = data.replace("): \n","):\n")
			elif "):  \n" in data: data = data.replace("):  \n","\n")
			elif "):   \n" in data: data = data.replace("):   \n","):\n")
			elif "):    \n" in data: data = data.replace("):    \n","):\n")
			elif "):     \n" in data: data = data.replace("):     \n","):\n")
			elif "):      \n" in data: data = data.replace("):      \n","):\n")
			elif "):       \n" in data: data = data.replace("):       \n","):\n")
			elif "):        \n" in data: data = data.replace("):        \n","):\n")
			#elif "\n " in data: data = data.replace("\n ", "\n")
			elif " \n" in data: data = data.replace(" \n", "\n")
			elif "	\n" in data: data = data.replace("	\n", "\n")
			elif "DOCS =  " in data: data = data.replace("DOCS =  ", "DOCS = ")
			elif "DOCS  = " in data: data = data.replace("DOCS  = ", "DOCS = ")
			elif "DOCS   = " in data: data = data.replace("DOCS   = ", "DOCS = ")
			elif "DOCS    = " in data: data = data.replace("DOCS    = ", "DOCS = ")
			elif "DOCS     = " in data: data = data.replace("DOCS     = ", "DOCS = ")
			else: break
		
		# slice with all indent (less dependable).
		if indent == None:
			to_slice = []
			for i in range(100):
				l = "\n"
				for _ in range(i): l += " "
				l += "def "
				to_slice.append(l)
			slices = String(data).split(to_slice)

		# slice by ident line.
		else:
			slices = data.split("\n"+String().indent(indent=indent)+"def ")

		# handle slices.
		if len(slices) == 1:
			return []
		elif len(slices) > 1:
			slices = slices[1:]
		slice_count = 0

		
		# iterate.
		for class_ in slices:
			if class_ not in [""," "]:
				type = "function"
				try: previous = slices[slice_count-1]
				except IndexError: previous = None
				if previous == None:
					init_line_indent = 0
				else:
					init_line_indent = 0 
					l = len(previous)
					for _ in range(len(previous)):
						l -= 1
						if previous[l] != " ": break
						init_line_indent += 1
					if init_line_indent > 0: init_line_indent += 1 # beause of the space after [class ].
				if previous != None and "@property" in previous:
					type = "property"
				if indent == None or str(init_line_indent) == str(int(indent)):
					class_ = String("def "+class_)
					class_line = class_.split("\n")[0]
					class_name = class_.split("(")[0].replace("def ","")
					if module != None:
						if module[len(module)-1] == ".": module = module[:-1]
						class_name = f"{module}.{class_name}"
					if func_allowed(class_name) and func_allowed(class_name, strip=True):
						class_parameters = String(class_).slice_tuple(depth=1)
						l_class = class_.replace(class_parameters+":","")
						for attempt in range(100):
							try:
								class_indent = String(l_class).line_indent(line=l_class.split("\n")[attempt])
								if class_indent != init_line_indent: break
							except IndexError:
								raise IndexError(f"Unable to determine the class indent of code: (\n{str(l_class)}\n).")
						if "\n" in class_parameters:
							class_code = class_.slice_indent(indent=class_indent, depth=2, remove_indent=not keep_indent)
							raw_class_code = class_.slice_indent(indent=class_indent, depth=2, remove_indent=False)
						else:
							class_code = class_.slice_indent(indent=class_indent, depth=1, remove_indent=not keep_indent)
							raw_class_code = class_.slice_indent(indent=class_indent, depth=1, remove_indent=False)
						classes, l_functions  = [], []
						if inside_classes and (("\nclass " in class_code or " class " in class_code) and class_code != data and len(class_code) > 0):
							l_module = module + "." + class_name
							classes = self.slice_classes(data=class_code, keep_indent=keep_indent, inside_functions=False, inside_classes=inside_classes, system_functions=system_functions, banned_functions=banned_functions, banned_classes=banned_classes, module=l_module)
							for i in classes:
								full_inside_class = f"{i['init_line'].split('(')[0]}{i['class_type']}:\n{i['raw_code']}"
								class_code = class_code.replace("):  \n","):\n").replace("):   \n","):\n").replace("):    \n","):\n").replace("):     \n","):\n").replace("):	\n","):\n").replace(full_inside_class, "")
								#class_code = class_code.replace(i["init_line"].split("(")[0]+class_parameters+":\n"+i["code"], "")
						if inside_functions and (("\ndef " in class_code or " def " in class_code) and class_code != data and len(class_code) > 0):
							l_module = module + "." + class_name
							l_functions = self.slice_functions(data=class_code, keep_indent=keep_indent, inside_classes=False, inside_functions=False, system_functions=system_functions, banned_functions=banned_functions, banned_classes=banned_classes, module=l_module)
							for i in l_functions:
								full_inside_class = f"{i['init_line'].split('(')[0]}{i['parameters']}:\n{i['raw_code']}"
								class_code = class_code.replace("):  \n","):\n").replace("):   \n","):\n").replace("):    \n","):\n").replace("):     \n","):\n").replace("):	\n","):\n").replace(full_inside_class, "")
						return_ = None
						if (" return response" in str(raw_class_code) or "\nreturn response" in str(raw_class_code)) or (" return dev0s.response." in str(raw_class_code) or "\nreturn dev0s.response." in str(raw_class_code)) or (" return _response_." in str(raw_class_code) or "\nreturn _response_." in str(raw_class_code)):
							return_ = "response"
						elif " return " not in str(raw_class_code) or "\nreturn " not in str(raw_class_code):
							return_ = "_"
						else:
							variables = raw_class_code.split(" return ")[len(raw_class_code.split(" return "))-1].split("\n")[0].replace(" ","").replace(",",", ")
							formats = []
							for variable in variables.split(", "):
								format = parse_variable_format(variable, class_parameters+":\n"+raw_class_code)
								if format == None:
									if skip_errors:
										format = "..."
									else:
										raise ValueError(f"Unable to determine the return format (#1) for function [{class_name}], code: [{class_code}]")
								formats.append(format)
							if len(formats) <= 0:
								if skip_errors:
									return_ = "..."
								else:
									raise ValueError(f"Unable to determine the return format (#2) for function [{class_name}], code: [{class_code}]")
							elif len(formats) <= 1:
								return_ = formats[0]
							else:
								return_ = Array(formats).string(joiner=", ")
						raw_name = str(class_name)
						if "." in raw_name: raw_name = raw_name.split('.')[len(raw_name.split('.'))-1]
						functions.append({
							"name":str(class_name),
							"raw_name":str(raw_name),
							"code":str(class_code),
							"raw":class_,
							"raw_code":str(raw_class_code),
							"full_code":f"{class_line.split('(')[0]}{class_parameters}:\n{class_code}",
							"raw_full_code":f"{class_line.split('(')[0]}{class_parameters}:\n{raw_class_code}",
							"indent":int(class_indent),
							"init_line":str(class_line),
							"parameters":str(class_parameters),
							"classes":classes,
							"functions":l_functions,
							"type":type,
							"return":return_,
						})
			slice_count += 1

		# fill DOCS customization.
		_functions_ = []
		for info in functions:
			initialized, module, description, chapter = False, None, [], None
			if "__init__" not in info["name"]:
				test = "DOCS = {"
				if test in info["code"]:
					customization = ast.literal_eval(str(String(info["code"].split(test[:-1])[1]).slice_dict(depth=1)))
					if "initialized" in customization:
						info["initialized"] = customization["initialized"]
					if "module" in customization:
						info["module"] = customization["module"]
					if "description" in customization:
						info["description"] = customization["description"]
					if "chapter" in customization:
						info["chapter"] = customization["chapter"]
					if "return" in customization:
						info["return"] = customization["return"]
					break
			for key,value in {
				"initialized":False,
				"module":None,
				"description":[],
				"chapter":None,
				"return":"_",
			}.items():
				try:info[key]
				except: info[key] = value
			_functions_.append(info)
		functions = _functions_

		# handler.
		return functions

		#

	# build code examples.
	def build_code_examples(self,
		# the path (leave None to use self.fp.path).
		path=None,
		# the package name (optional).
		package=None,
		# the root path in case the path is a dir (optional).
		root=None,
		# include system functions (aka: def __somefunc__).
		system_functions=False,
		# banned sub paths.
		banned=[],
		# banned names.
		banned_names=["__main__.py", "utils.py", ".version"],
		# banned basenames.
		banned_basenames=["utils", "__pycache__", ".legacy"],
		# banned class types.
		banned_class_types=["Exception"],
		# the banned classes.
		banned_classes=[],
		# the banned functions.
		banned_functions=[],
		# return format (str / dict) (always returns dict if parameter path is a list).
		format=str,
		# readme enabled.
		readme=False,
		# the log level.
		log_level=defaults.options.log_level,
		# get all ids (should not be used).
		get_all_ids=False,
	):

		# funcs.
		def clean_params(params):
			clean = params.replace('(self)', '()').replace('self  ,', '').replace('self ,', '').replace('self,', '').replace("\n)", " )")
			while True:
				if clean[:len("( ")] == "( ":
					clean = "("+clean[2:]
				else: break
			return clean
		def build_class_doc(info):
			unique_name = str(info["raw_name"])
			info['raw_name'] = info['raw_name'].split("-")[0]
			if "." in info["raw_name"]: l = info["raw_name"].split(".")[len(info["raw_name"].split("."))-1]
			else: l = info["raw_name"]
			initialized_name = String(l).variable_format()
			old_init_name = initialized_name
			#class_holder_class = True
			#if len(initialized_name) >= 2 and String(initialized_name).first("__") == "__":
			#	initialized_name = String(initialized_name).remove_first("__")
			#else: class_holder_class = False
			#if len(initialized_name) >= 2 and String(initialized_name).last("__") == "__":
			#	initialized_name = String(initialized_name).remove_last("__")
			#else: class_holder_class = False
			#if class_holder_class:
			#	initialized_name = String(initialized_name).capitalized_word()
			""" """

			# fill DOCS customization.
			initialized, module, description = info["initialized"], info["module"], info["description"]
			if module == None:
				module = info["name"]

			# docs.
			parameters = clean_params(info["parameters"]).replace("self.", initialized_name+".")
			if initialized == True:
				doc = "\n"
				if description != None:
					for i in description: doc += f"{i}\n".replace("</attribute>", "").replace("<attribute>", "").replace("</parameter>", "").replace("<parameter>", "")
				doc += (
					"\n# import the " + str(module) + " object class." + "\n" +
					"")
				doc += (
					#f"from {package} import " + module.split('.')[0] + "\n" +
					f"import {package}" + "\n" +
					"")
			else:
				doc = "\n"
				if description != None:
					for i in description: doc += f"{i}\n".replace("</attribute>", "").replace("<attribute>", "").replace("</parameter>", "").replace("<parameter>", "")
				doc += (
					"\n# initialize the " + str(module) + " object class." + "\n" +
					"")
				doc += (
					str(initialized_name)+" = " + module + str(parameters) + "\n" +
					"")
			if readme:
				doc = (
					f"## {info['raw_name']}:" + "\n" +
					f"The {initialized_name} object class." + "\n" +
					f"``` python " + "\n" +
					doc + "\n" +
					f"```" + "\n" +
					"")
			properties = []
			c = 0
			for i in info["functions"]:
				if "__init__" not in i["name"]:
					#if module != None:
					#	i["name"] = i["name"].replace(f"{old_init_name}.", f"{module}.")
					#	i["raw_name"] = i["raw_name"].replace(f"{old_init_name}.", f"{module}.")
					if "." in i["name"]:
						l = i["name"].split(".")[len(i["name"].split("."))-1]
					else:
						l = i["name"]
					i["name"] = f'{initialized_name}.{l}'
					#i["raw_name"] = f'{initialized_name}.{l}'
					i["parameters"] = clean_params(i["parameters"]).replace("self.", initialized_name+".") # just for the self / initialized_name replacement.
					if i["type"] == "property":
						properties.append(i)
					else:
						if c == 0:
							doc += "\n#### Functions:" + "\n\n"
						c += 1
						doc += build_func_doc(i, small_readme=True)
			if c > 0: doc += "\n"
			if len(properties) > 0: 
				doc += "#### Properties:" + "\n"
				for i in properties: 
					if "__init__" not in i["name"]:
						doc += build_property_doc(i)
				doc += "\n"
			return doc
		def build_func_doc(info, small_readme=False):
			unique_name = str(info["raw_name"])
			info['raw_name'] = info['raw_name'].split("-")[0]
			parameters = clean_params(info["parameters"])
			doc = (
				"\n# call " + info["name"] + "." + "\n" +
				f"{info['return']} = " + info['name'] + parameters + "\n" +
				"")
			if "__init__" in info["name"]:
				return ""
			if readme:
				if small_readme:
					doc = (
						f"##### {info['raw_name']}:" + "\n" +
						f"``` python" + "\n" +
						doc + "\n" +
						f"```" + "\n" +
						"")
				else:
					doc = (
						f"#### {info['raw_name']}:" + "\n" +
						f"The {info['name']} function." + "\n" +
						f"``` python" + "\n" +
						doc + "\n" +
						f"```" + "\n" +
						"")
			return doc
		def build_property_doc(info):
			full_name = str(info["name"])
			if "." in info["name"]:
				info["name"] = info["name"].split(".")[1]
			doc = (
				"```python" + "\n"
				"\n# the " + info["name"].replace("_"," ") + " property." + "\n" +
				info["name"]+" = " + full_name + "\n" +
				"```" + "\n"
				"")
			return doc
			#

		# slice.
		if path == None: path = self.fp.path
		if path == None:
			raise dev0s.exceptions.InvalidUsage(f"Define parameter [path] (str).")
		if gfp.exists(path=path) and gfp.directory(path=path):
			path = Directory(path).paths(extensions=["py"], banned=banned, banned_names=banned_names, banned_basenames=banned_basenames, recursive=True)
		if isinstance(path, (str, String)): path = [path]
		if root != None: root = gfp.clean(path=root)
		docs_dict = {}
		for _path_ in path:

			# load.
			if log_level >= 0:
				loader = console.Loader(f"Building code examples for [{_path_}].")
			data = Files.load(_path_)
			module = None
			if root != None: 
				module = gfp.module(path=_path_.replace(root, ""))
				if package != None:
					module = package+"."+module
			first_indent_functions = self.slice_functions(indent=0, module=module, data=data, system_functions=system_functions, banned_functions=banned_functions, banned_classes=banned_classes)
			classes = self.slice_classes(data=data, module=module, system_functions=system_functions, banned_functions=banned_functions, banned_classes=banned_classes)

			# iterate.
			for info in first_indent_functions:
				if "__init__" not in info["name"]:
					try: info["chapter"]
					except KeyError: info["chapter"] = "Functions"
					if info["chapter"] == None: info["chapter"] = "Functions"
					try: docs_dict[info["chapter"]]
					except KeyError: docs_dict[info["chapter"]] = {}
					if info["raw_name"] in list(docs_dict[info["chapter"]].keys()):
						try:info["raw_name"] = info["raw_name"].split("-")[0]+"-"+str(int(info["raw_name"].split("-")[1])+1)
						except: info["raw_name"] = info["raw_name"]+"-1"
					docs_dict[info["chapter"]][info["raw_name"]] = {
						"docs":build_func_doc(info),
						"name":info["name"],
						"raw_name":info["raw_name"].split("-")[0],
						"unique_name":info["raw_name"],
						"type":info["type"],
						#
						"parameters":info["parameters"],
						"initialized":info["initialized"],
						"module":info["module"],
						"description":info["description"],
						"chapter":info["chapter"],
						"return":info["return"],
						"path":_path_,
					}
			for info in classes:
				l_banned = False
				for i in banned_class_types:
					if i in info["class_type"].split(", "):
						l_banned = True
						break
				if not l_banned:
					try: info["chapter"]
					except KeyError: info["chapter"] = "Classes"
					if info["chapter"] == None: info["chapter"] = "Classes"
					try: docs_dict[info["chapter"]]
					except KeyError: docs_dict[info["chapter"]] = {}
					if info["raw_name"] in list(docs_dict[info["chapter"]].keys()):
						try:info["raw_name"] = info["raw_name"].split("-")[0]+"-"+str(int(info["raw_name"].split("-")[1])+1)
						except: info["raw_name"] = info["raw_name"]+"-1"
					functions = []
					for func in info["functions"]:
						func["path"] = _path_
						functions.append(func)
					info["functions"] = functions
					docs_dict[info["chapter"]][info["raw_name"]] = {
						"docs":build_class_doc(info),
						"name":info["name"],
						"raw_name":info["raw_name"].split("-")[0],
						"unique_name":info["raw_name"],
						"type":info["type"],
						"functions":info["functions"],
						#
						"parameters":info["parameters"],
						"initialized":info["initialized"],
						"module":info["module"],
						"description":info["description"],
						"chapter":info["chapter"],
						"path":_path_,
					}
			if log_level >= 0: loader.stop()
		
		# sort alphabetically.
		docs_dict = Dictionary(docs_dict).sort(alphabetical=True)

		# assign unique id's.
		_docs_dict_ = {}
		all_ids = []
		for chapter, l_docs_dict in docs_dict.items():
			try: _docs_dict_[chapter]
			except KeyError: _docs_dict_[chapter] = {}
			for _, info in l_docs_dict.items():
				while True:
					if info["raw_name"] in all_ids:
						try:info["raw_name"] = info["raw_name"].split("-")[0]+"-"+str(int(info["raw_name"].split("-")[1])+1)
						except: info["raw_name"] = info["raw_name"]+"-1"
					else:
						all_ids.append(info["raw_name"])
						if info["type"] == "class":
							new = []
							for _info_ in info["functions"]:
								if "__init__" not in _info_["name"]:
									while True:
										if _info_["raw_name"] in all_ids:
											#print(_info_["name"], "vs", all_ids)
											try:_info_["raw_name"] = _info_["raw_name"].split("-")[0]+"-"+str(int(_info_["raw_name"].split("-")[1])+1)
											except: _info_["raw_name"] = _info_["raw_name"]+"-1"
										else:
											all_ids.append(_info_["raw_name"]) 
											break
									new.append(_info_)
							info["functions"] = new
						_docs_dict_[chapter][info["raw_name"]] = info
						break
		docs_dict = _docs_dict_

		# return.
		if format == str and len(docs_dict) == 1:
			if get_all_ids:
				return docs_dict[_path_], all_ids
			else:
				return docs_dict[_path_]
		else:
			if get_all_ids:
				return docs_dict, all_ids
			else:
				return docs_dict

		#

	# build readme code examples.
	def build_readme(self,
		# the package name.
		package="dev0s",
		# the root path in case the path is a dir (optional).
		root=None,
		# the package's description (leave "" or None to ignore).
		description="",
		# the package's install command (leave "" or None to ignore).
		installation="pip3 install dev0s --upgrade",
		# include system functions (aka: def __somefunc__).
		system_functions=False,
		# include different path's or readme str.
		include=[],
		# banned sub paths.
		banned=[],
		# banned names.
		banned_names=["__main__.py", "utils.py", ".version"],
		# banned basenames.
		banned_basenames=["utils", "__pycache__", ".legacy"],
		# banned class types.
		banned_class_types=["Exception"],
		# the banned classes.
		banned_classes=[],
		# the banned functions.
		banned_functions=[],
		# the readme export path (leave None to ignore).
		readme=None,
		# the examples export path (leave None to ignore).
		examples=None,
		# the log level.
		log_level=defaults.options.log_level,
		# the post readme replacements.
		replacements={},
	):
		
		# header.
		_readme_ =  ""
		header = (
			f"""# {package}""" + "\n" + 
			"""Author(s):  Daan van den Bergh.<br>""" + "\n" + 
			"""Copyright:  © 2020 Daan van den Bergh All Rights Reserved.<br>""" + "\n" + 
			"""Supported Operating Systems: macos & linux.<br>""" + "\n" + 
			"""<br>""" + "\n" + 
			"""<br>""" + "\n" + 
			"""<p align="center">""" + "\n" + 
			"""  <img src="https://raw.githubusercontent.com/vandenberghinc/public-storage/master/vandenberghinc/icon/icon.png" alt="Bergh-Encryption" width="50"> """ + "\n" + 
			"""</p>""" + "\n" + 
			"")

		# description.
		if description not in ["", None]:
			description = description.replace("\n", "<br>")
			_readme_ += (
				f"""\n# Description:""" + "\n" + 
				f"""{description}""" + "\n" + 
				"")

		# installation.
		if installation not in ["", None]:
			installation = installation.replace("\n", "<br>")
			_readme_ += (
				f"""\n# Installation:""" + "\n" + 
				f"""Install the package.""" + "\n" + 
				f"""\n	{installation}""" + "\n" + 
				"")

		# include.
		if include not in ["", None]:
			if isinstance(include, str):
				if os.path.exists(include):
					include = Files.load(include)
				_readme_ += (
					f"""\n{include}""" + "\n" + 
					"")
			elif isinstance(include, (list,Array)):
				for i in include:
					if os.path.exists(i):
						i = Files.load(i)
					_readme_ += (
						f"""\n{i}""" + "\n" + 
						"")

		# code examples.
		code_examples, all_ids = self.build_code_examples(format=dict, readme=True, get_all_ids=True, system_functions=system_functions, banned=banned, banned_names=banned_names, banned_basenames=banned_basenames, banned_class_types=banned_class_types, banned_classes=banned_classes, banned_functions=banned_functions, log_level=log_level, root=root, package=package)
		_code_examples_ = ""
		if len(code_examples) > 0:
			l = (
				f"""\n# Code Examples:""" + "\n" + 
				"")
			
			# add table content.
			l += f"""\n### Table of content:""" + "\n"
			last_chapter = None
			for chapter, l_code_examples in code_examples.items():
				for id, example in l_code_examples.items(): 
					if example["type"] == "class":
						if chapter == None: chapter = "Classes"
						if chapter != last_chapter:
							l += f"- [__{chapter}__](#{chapter.replace(' ','-').lower()})" + "\n"
							last_chapter = chapter
						l += f"  - [__{example['raw_name']}__](#{example['unique_name'].replace(' ','-').lower()})" + "\n"
						properties = False
						for fexample in example["functions"]:
							if "__init__" not in fexample["raw_name"]: 
								if fexample["type"] in ["property"]:
									if not properties:
										properties = True
										unique = "properties"
										while True:
											if unique not in all_ids:
												all_ids.append(unique)
												break
											else:
												try:unique = unique.split("-")[0]+"-"+str(int(unique.split("-")[1])+1)
												except: unique = unique+"-1"
										unique_name = str(unique)
										raw_name = str(fexample['raw_name']).split("-")[0]
										l += f"    * [{raw_name}](#{unique_name.replace(' ','-').lower()})" + "\n"	
								else:
									unique_name = str(fexample['raw_name'])
									raw_name = str(fexample['raw_name']).split("-")[0]
									l += f"    * [{raw_name}](#{unique_name.replace(' ','-').lower()})" + "\n"
			l += "\n"#------\n\n"
			_code_examples_ += l
			_readme_ += l
			
			# add codes.
			for chapter, l_code_examples in code_examples.items():
				for _, example in l_code_examples.items(): 
					#print(example)
					_code_examples_ += example["docs"]
					_readme_ += example["docs"]

		# create header table of content.
		code_open, headers = False, []
		for line in _readme_.split("\n"):
			if '```' in line: 
				if code_open: code_open = False
				else: code_open = True
			if not code_open and len(line) >= len("# ") and line[:len("# ")] == "# ":
				h = line[len("# "):].replace(":","")
				while True:
					if len(h) > 0 and h[len(h)-1] == " ": h = h[:-1]
					elif len(h) > 0 and h[0] == " ": h = h[1:]
					else: break
				headers.append(h)
		table = f"""\n## Table of content:""" + "\n"
		for name in headers: 
			table += f"  * [{name}](#{name.replace(' ','-').lower()})" + "\n"
		table += "\n"#------\n\n"
		_readme_ = header + table + _readme_

		# normalize.
		_readme_ = _readme_.replace("\n\n\n","\n\n").replace("\n\n\n","\n\n").replace("\n\n\n","\n\n").replace("\n\n\n","\n\n")
		_code_examples_ = _code_examples_.replace("\n\n\n","\n\n").replace("\n\n\n","\n\n").replace("\n\n\n","\n\n").replace("\n\n\n","\n\n")

		# replacements must be at the end so the user can test and add replacements.
		for from_, to_ in replacements.items():
			_readme_ = _readme_.replace(from_, to_)
			_code_examples_ = _code_examples_.replace(from_, to_)

		# readme export.
		if readme != None:
			if log_level >= 0: loader = console.Loader(f"Saving created readme to [{readme}]")
			if not os.path.exists(gfp.base(path=readme)): os.system(f"mkdir -p {gfp.base(path=readme)}")
			Files.save(readme, _readme_)
			if log_level >= 0: loader.stop()

		# examples export.
		if examples != None:
			if log_level >= 0: loader = console.Loader(f"Saving created code examples to [{examples}]")
			if not Files.exists(gfp.base(examples)): Files.create(gfp.base(examples), directory=True)
			Files.save(examples, _code_examples_)
			if log_level >= 0: loader.stop()

		# handler.
		return code_examples, _readme_
		
		#

	# clean python code.
	def clean(self,
		# forced enabled ignored the warning prompt (bool).
		forced=False,
		# the self path (str) (leave None to use self.path).
		path=None,
	):

		raise ValueError("Coming soon.")

		# prompt warning.
		if not forced and not console.input(f"&RED&Warning&END&: You are cleaning python code [{path}], do you wish to proceed?", yes_no=True):
			raise Exceptions.AbortError("User aborted.")

		# vars.
		if path == None: path = self.fp.path
		if path == None: raise Exceptions.InvalidUsage("<Python.clean> define parameter: [path].")

		""" load data """
		data = Files.load(path=path, format="str")

		""" normalize data """
		while True:
			if "	" in data: data = data.replace("	", "    ")
			else: break

		"""
		add # to the the end of each indentation.
		"""
		def check_indentation_end(data):
			new_data, string = "", String()
			items = String(data).iterate_lines()
			max = len(items)-1
			last_indent = None
			last_chars = {
				"1":"",
				"blank":False,
			}
			for linecount, line in items:
				indent = string.line_indent(line=line)
				print(f"{indent}:[{line}]")
				if last_indent != None and last_chars["1"] != "#" and indent < last_indent:
					print("Found lower indent difference")
					difference = indent - last_indent
					new_data += string.indent(indent=last_indent)+"#\n"
				if linecount == max: new_data += line
				else: new_data += line+"\n"
				last_indent = indent
				if len(line) > 0: l = line[len(line)-1]
				else: l = ""

				# add last chars.
				last_chars["blank"] = line in [""]
				for key, _ in last_chars.items():
					if key not in ["blank"]:
						last_chars[key] += l
						while True:
							if len(last_chars[key]) >= int(key): last_chars[key] = last_chars[key][:-1]
							else: break

				#

			return new_data

		# call cleanings.
		old = str(data)
		data = check_indentation_end(data)

		# results.
		print("\n=====================================\nFrom:\n"+old)
		print("\n=====================================\nTo:\n"+data)


	# object instance.
	def instance(self):
		return "Python"
	@property
	def __name__(self):
		return self.instance()
		
		#

	# support self assignment.
	def assign(self, string):
		if isinstance(data, (int, float)):
			a=1
		elif isinstance(data, (String)):
			data = data.string
		elif isinstance(data, self.__class__):
			data = data.data
		elif not isinstance(data, self.__class__):
			raise Exceptions.FormatError(f"Can not assign object {self.__class__} & {data.__class__}.")
		self.data = str(data)
		return self
		
		#

	#

# the legacy class.
class legacy:
	def export(
		# the absolute source path.
		source=None,
		# the source directory / file names that will be included in the export.
		names = [
			"apps",
			"classes",
			"models",
			"requirements",
			"scripts",
			"NOTES.md",
			"README.md",
			"__main__.py",
			"manage.py",
			"__defaults__",
			".version",
			# can also be like this.
			#".cache/weights/",
		],
	):

		# export.
		if source == None or not Files.exists(source): raise FileNotFoundError(f"Unable to find source directory {source}.")
		version = Files.load(Files.join(source, ".version")).replace("\n","").replace(" ","")
		export = FilePath(Files.join(source, f".legacy/{version}/"))
		if export.exists(): raise ValueError(f"Legacy export path {export} already exists.")
		for name in names:
			from_, to_ = Files.join(source, name), Files.join(export, name) 
			Files.copy(from_, to_)
			print(f"Included {from_}")
		new_version = String(version).increase_version()
		Files.save(Files.join(source, ".version"), new_version)
		print(f"Successfully legacy exported version {version}.")
		print(f"Incremented source version to {new_version}.")

#




# imports
from dev0s.classes.config import *

# selenium.
from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium import webdriver


# the disks object class.
class Browser(object):
	def __init__(self,
		# the driver.
		driver="chromedriver",
	):

		# docs.
		DOCS = {
			"module":"dev0s.system.Browser", 
			"initialized":False,
			"description":[], 
			"chapter": "System", }
		
		# objects.
		#options = Options()
		#options.binary_location = "C:\\Program Files\\Chrome\\chrome64_55.0.2883.75\\chrome.exe"
		#driver = webdriver.Chrome(chrome_options = options, executable_path=r'C:\path\to\chromedriver.exe')
		if driver in ["chrome", "google chrome", "chromedriver"]:
			#self.driver = webdriver.Chrome(f'{SOURCE_PATH}/classes/browser/drivers/{OS}/chromedriver')
			self.driver = webdriver.Chrome("/usr/local/bin/chromedriver")
		elif driver in ["firefox", "geckodriver"]:
			#self.driver = webdriver.Firefox(executable_path=f'{SOURCE_PATH}/classes/browser/drivers/{OS}/geckodriver')
			self.driver = webdriver.Firefox(executable_path="/usr/local/bin/geckodriver")
		else:
			raise ValueError(f"Unknown driver: {driver}")

		#
	def get(self, url):
		self.driver.get(url)
	def get_element(self,
		# the element type.
		element="input",
		# the attribute name.
		attribute="name",
		# the attributes value.
		value="username",
		# the parent element (default is self.driver).
		parent=None, 
	):
		#search_bar = driver.find_element_by_name("q")
		if parent == None:
			return self.driver.find_element_by_xpath(f"//{element}[@{attribute}='{value}']")
		else:
			return parent.find_element_by_xpath(f".//{element}[@{attribute}='{value}']")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
import os, sys, requests, ast, json, pathlib, platform, random, shutil, zipfile, pwd, grp, subprocess, time, threading, sysconfig, urllib, flask, multiprocessing, socket
from datetime import datetime

# settings.
ALIAS = "dev0s"

# update.
if "--update" in sys.argv and ALIAS in sys.argv[0]:
	os.system(f"curl -s https://raw.githubusercontent.com/vandenberghinc/{ALIAS}/master/{ALIAS}/requirements/installer.remote | bash ")
	sys.exit(0)

# functions.
def __get_operating_system__():
	os = platform.system().lower()
	if os in ["darwin"]: return "macos"
	elif os in ["linux"]: return "linux"
	else: raise ValueError(f"Unsupported operating system: [{os}].")
def __get_source_path__(package_name, index=1):
	executable_dir = str(pathlib.Path(__file__).absolute()).replace(os.path.basename(pathlib.Path(__file__)), '').replace("//","/")
	if executable_dir[len(executable_dir)-1] == "/": executable_dir = executable_dir[:-1]
	source, c = "/", 1
	for id in executable_dir.split("/"):
		if id == package_name:
			if c == index:
				source += id+"/"
				break
			else: c += 1
		else: source += id+"/"
	base = source[:-1].split("/")
	base = source.replace(f'/{base[len(base)-1]}/', '/')
	return source, base
def __save_file__(path, data):
	file = open(path, "w+") 
	file.write(data)
	file.close()
def __load_file__(path, data="not to be used", format="str"): # keep data as second param to prevent save load errors.
        # correct format.
        if format in ["string", "file"]: format = "str"
        if format in ["dict", "array"]: format = "json"
        # match format.
        if format == "str":
            file = open(path,mode='rb')
            data = file.read().decode()
            file.close()
            return data
        elif format == "json":
            data = None
            with open(path, "r") as json_file:
                data = json.load(json_file)
            return data
        elif format == "bytes":
            data = None
            with open(path, "rb") as file:
                data = file.read()
            return data
        else: raise ValueError(f"Unknown format {format}.")

# source.
SOURCE_PATH, BASE = __get_source_path__(ALIAS, index=1)
OS = __get_operating_system__()

# file settings.
USER = OWNER = os.environ.get("USER")
GROUP = "root"
HOME = os.environ.get('HOME')+"/"
if OS in ["macos"]:  GROUP = "wheel"

# options.
ERROR_TRACEBACK = os.environ.get("R3SPONSE_ERROR_TRACEBACK")
if str(ERROR_TRACEBACK) in ["True", "true", True]: ERROR_TRACEBACK = True
else: ERROR_TRACEBACK = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.config import *

# the console.* functions (exact copy).
def argument_present(arguments, default=False, count=1):
	if isinstance(arguments, str):
		c = sys.argv.count(arguments)
		if c > 0 and count <= c:
			return True
		else: return default
	elif isinstance(arguments, list):
		for argument in arguments:
			c = sys.argv.count(argument)
			if c > 0 and count <= c:
				return True
		return default
	else: raise ValueError("Invalid usage, arguments must either be a list or string.")
def arguments_present(arguments, default=False, count=1):
	if isinstance(arguments, str): return argument_present(arguments, default=default, count=count)
	else:
		for argument in arguments:
			if argument_present(argument, default=default, count=count):
				return True
		return default
def get_argument(argument, required=True, index=1, count=1, default=None, ):
	if default != None and required: required = False
	
	# set error msg.
	if index == 1:
		empty_error = f"Define argument [{argument}]."
	else:
		empty_error = f"Define argument [{argument}] (index: {index})."

	# check presence.
	if argument not in sys.argv:
		if required:
			raise EmptyArgumentError(empty_error)
		else: return default

	# retrieve.
	y, c = 0, 0
	for x in sys.argv:
		try:
			if x == argument: 
				c += 1
				if count == c:
					return sys.argv[y+index]
		except IndexError:
			if required:
				raise EmptyArgumentError(empty_error)
			else: return default
		y += 1

	# should not happen.
	if required:
		raise EmptyArgumentError(empty_error)
	else: return default

	#





# imports
from dev0s.classes.config import *
from dev0s.classes.defaults.objects import *
from dev0s.classes.defaults.defaults import defaults
from dev0s.classes.response import response as _response_

# utils.
def check_os(supported=["linux"], error=False):
	if OS not in supported:
		if error: raise ValueError(f"Unsupported operating system: {OS}.")
		else: return _response_.error(f"Unsupported operating system: {OS}.")
	else:
		if error: return None
		else: return _response_.success(f"Supported operating system: {OS}.")
def coming_soon():
	raise ValueError("COMING SOON.")

# the disks object class.
class Disks(Object):
	def __init__(self):
		
		# docs.
		DOCS = {
			"module":"dev0s.system.disks", 
			"initialized":True,
			"description":[], 
			"chapter": "System", }

		# defaults.
		Object.__init__(self, traceback="dev0s.system.disks")

		#
	def list(self):
		coming_soon()

		# check os.
		response = check_os()
		if not response.success: return response

		#
	def erase(self, 
		# the device without partition number (/dev/sdb).
		device=None,
	):

		# check parameters.
		response = _response_.parameters.check({
			"device:str":device,})
		if not response.success: return response

		# check os.
		if defaults.vars.os not in ["linux"]:
			return _response_.error(f"Unsupported operating system [{defaults.vars.os}].")

		# handler.
		if "SUCCESS!" in output:
			return _response_.success(f"Successfully erase device {device}.")
		else:
			return _response_.error(f"Failed to erase device {device}, output: {output}")

		#
	def partition(self, 
		# the device without partition number (/dev/sdb).
		device=None,
	):
		coming_soon()
		# check parameters.
		response = _response_.parameters.check({
			"device:str":device,})
		if not response.success: return response

		# check os.
		response = check_os()
		if not response.success: return response

		# execute.
		output = utils.__execute_script__(f"""
	        sudo parted {device} mklabel gpt
	        sudo parted -a opt {device} mkpart primary ext4 0% 100%
	        """)
		
		# handler.
		if "SUCCESS!" in output:
			return _response_.success(f"Successfully partitioned device {device}.")
		else:
			return _response_.error(f"Failed to partition device {device}, output: {output}")

		#
	def format(self, 
		# the device with partition number (/dev/sdb1).
		device=None,
		# the assigned label (name).
		label=None,
	):
		coming_soon()
		# check parameters.
		response = _response_.parameters.check({
			"device:str":device,
			"label:str":label,})
		if not response.success: return response
		
		# check os.
		response = check_os()
		if not response.success: return response

		# handler.
		if "SUCCESS!" in output:
			return _response_.success(f"Successfully formatted device {device}.")
		else:
			return _response_.error(f"Failed to format device {device}, output: {output}")

		#
	def mount(self, 
		# the device with partition number (/dev/sdb1).
		device=None, 
		# the mountpoint path.
		path=None,
	):
		coming_soon()
		# check parameters.
		response = _response_.parameters.check({
			"device:str":device,
			"path:str":path,})
		if not response.success: return response
		
		# check os.
		response = check_os()
		if not response.success: return response

		# handler.
		if "SUCCESS!" in output:
			return _response_.success(f"Successfully mounted device {device} to {path}.")
		else:
			return _response_.error(f"Failed to mount device {device} to {path}, output: {output}")

		#
	def unmount(self, 
		# the mountpoint path.
		path=None,
	):
		coming_soon()
		# check parameters.
		response = _response_.parameters.check({
			"path:str":path,})
		if not response.success: return response
		
		# check os.
		response = check_os()
		if not response.success: return response

		# handler.
		if "SUCCESS!" in output:
			return _response_.success(f"Successfully unmounted path {path}.")
		else:
			return _response_.error(f"Failed to unmount path {path}, output: {output}")

		#

# initialized ojects.
disks = Disks()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# insert the package for universal imports.
def __get_source_path__(
	# the path (leave None to use self.path) (param #1).
	path=None,
	# the dirs back.
	back=1,
):
	base = path.replace('//','/')
	if base[len(base)-1] == '/': base = base[:-1]
	if len(base.split("/")) <= 1: raise ValueError("Path [{}] has no base.".format(base))
	startslash = True
	if base[0] != "/":
		startslash = False
	base = base.split("/")
	m, c, s = len(base), 0, ""
	for i in base:
		if c >= m-back: break
		if c == 0:
			s = f"/{i}/"
		else:
			s += f"{i}/"
		c += 1
	if startslash:
		return s
	else:
		return s[1:]
	###### OLD.
	base = path.replace('//','/')
	if len(base.split("/")) <= 1: raise ValueError("Path [{}] has no base.".format(base))
	if base[len(base)-1] == '/': base = base[:-1]
	for x in range(back, back+1):
		last = (base.split('/')[len(base.split('/'))-1]).replace('//','/')
		base = base[:-len("/"+last)]
	while True:
		if '//' in base: base = base.replace('//','/')
		else: break
	if base[len(base)-1] != "/": base += '/'
	return base
	"""splitted, result, count = path.split('/'), "", 0
	for i in splitted:
		if count < len(splitted) - 1 - back:
			result += '/' + i
		else: result += "/"
		count += 1
	"""
import sys ; sys.path.insert(1, __get_source_path__(__file__, back=2))

# imports.
from dev0s.classes.config import *
from dev0s.shortcuts import *
import getpass

# the cli object class.
class CLI(dev0s.cli.CLI):
	def __init__(self):
		
		# cli.
		dev0s.cli.CLI.__init__(self,
			modes={
				"Encryption":"*chapter*",
				"    --encrypt /path/to/input /path/to/output [optional: --remove]":"Encrypt the provided file path.",
				"    --decrypt /path/to/input /path/to/output [optional: --remove]":"Decrypt the provided file path.",
				"    --encrypt-env . [optional: --remove]":"Encrypt the specified enviroment (automatically fills: --key ./key --input ./data/ --output ./data.enc.zip).",
				"    --decrypt-env . [optional: --remove]":"Decrypt the specified enviroment (automatically fills: --key ./key --input ./data.enc.zip --output ./data/).",
				"    --generate-keys":"Generate a key pair.",
				"    --generate-passphrase [optional: --length 32]":"Generate a passphrase.",
				"    --generate-aes [optional: --length 64]":"Generate an aes passphrase.",
				"System":"*chapter*",
				"    --user myuser":"Configure a system user (linux).",
				"        --check":"Check the existance of the specified user.",
				"        --create":"Create the specified user.",
				"        --delete":"Delete the specified user.",
				"        --password MyPassword":"Set the password of the specifed user (leave password blank for safe prompt).",
				"        --add-groups group1,group2":"Add the specified user to groups.",
				"        --delete-groups group1,group2":"Remove the specified user from groups.",
				"    --group mygroup":"Configure a system group (linux).",
				"        --check":"Check the existance of the specified group.",
				"        --create":"Create the specified group.",
				"        --delete":"Delete the specified group.",
				"        --list-users":"List the users of the specified group.",
				"        --add-users user1,user2":"Add users from the specified group.",
				"        --delete-users user1,user2":"Delete users from the specified group.",
				"        --force-users user1,user2":"Add the specified users to the group and remove all others.",
				"    --disk-space":"Get the free disk space.",
				"    --size /path/to/file":"Get the size of a file / directory.",
				"Metrics":"*chapter*",
				"    --metrics":"Access the metrics.",
				"      --ram":"Parse the RAM metrics (linux).",
				"CleanUp":"*chapter*",
				"    --cleanup":"Access the cleanup.",
				"      --ram ":"Clean up the RAM memory (linux).",
				"      --swap ":"Clean up the RAM swap memory (linux).",
				"Network":"*chapter*",
				"    --network":"Access the network.",
				"        --info":"Retrieve the current network information.",
				"    --firewall":"Access the firewall (linux only).",
				"        --info":"Retrieve the current firewall information.",
				"        --disable":"Disable the firewall.",
				"        --enable":"Enable the firewall.",
				"        --set-default false":"Set the default firewall behaviour (deny/allow).",
				"        --allow 22":"Allow a port in the firewall settings.",
				"        --deny 22":"Deny a port in the firewall settings.",
				"Installation":"*chapter*",
				"    --install":f"Install the {ALIAS} library.",
				"    --uninstall":f"Uninstall the {ALIAS} library.",
				"    --reinstall":f"Reinstall the {ALIAS} library.",
				"    --link":f"Link (activate) the {ALIAS} library.",
				"    --unlink":f"Unlink (deactivate) the {ALIAS} library.",
				"    --update":f"Update the {ALIAS} library.",
				"Defaults":"*chapter*",
				"    --version":"Show the dev0s version.",
				"    -h / --help":"Show the documentation.",
			},
			options={
			},
			executable=__file__,
			alias=ALIAS,)

		#
	def start(self):
		
		# check arguments.
		self.arguments.check(exceptions=["--log-level", "--create-alias", "--version", "--non-interactive", "--remove"], json=dev0s.defaults.options.json)

		# ____________________________________________________________________________________________________
		#
		# BASICS
		#

		# help.
		if self.arguments.present(['-h', '--help']):
			self.docs(success=True, json=dev0s.defaults.options.json)

		# version.
		elif self.arguments.present(['--version']):
			print(f"{ALIAS} version:",Files.load(f"{SOURCE_PATH}/.version").replace("\n",""))

		# ____________________________________________________________________________________________________
		#
		# ENCRYPTION
		#

		# encrypt.
		elif self.arguments.present('--encrypt'):
			input = self.arguments.get('--encrypt', index=1, json=dev0s.defaults.options.json)
			output = self.arguments.get('--encrypt', index=2, json=dev0s.defaults.options.json)
			encryption = self.get_encryption(prompt_passphrase=False)
			response = encryption.load_public_key()
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			if not response.success: sys.exit(1)
			if os.path.isdir(input): 
				response = encryption.encrypt_directory(input=input, output=output, remove=self.arguments.present("--remove"))
			else: 
				response = encryption.encrypt_file(input=input, output=output, remove=self.arguments.present("--remove"))
			self.stop(response=response, json=dev0s.defaults.options.json)

		# decrypt.
		elif self.arguments.present('--decrypt'):
			input = self.arguments.get('--decrypt', index=1, json=dev0s.defaults.options.json)
			output = self.arguments.get('--decrypt', index=2, json=dev0s.defaults.options.json)
			encryption = self.get_encryption()
			response = encryption.load_private_key()
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			if not response.success: sys.exit(1)
			if os.path.isdir(input) or (".enc.zip" in input and ".enc.zip" not in output): 
				response = encryption.decrypt_directory(input=input, output=output, remove=self.arguments.present("--remove"))
			else: 
				response = encryption.decrypt_file(input=input, output=output, remove=self.arguments.present("--remove"))
			self.stop(response=response, json=dev0s.defaults.options.json)

		# encrypt env.
		elif self.arguments.present('--encrypt-env'):
			env = self.arguments.get('--encrypt-env', json=dev0s.defaults.options.json)
			key = f"{env}/key/".replace("//","/").replace("//","/").replace("//","/")
			input = f"{env}/data/".replace("//","/").replace("//","/").replace("//","/")
			output = f"{env}/data.enc.zip".replace("//","/").replace("//","/").replace("//","/")
			encryption = self.get_encryption(prompt_passphrase=False, key=key)
			response = encryption.load_public_key()
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			if not response.success: sys.exit(1)
			response = encryption.encrypt_directory(input=input, output=output, remove=self.arguments.present("--remove"))
			self.stop(response=response, json=dev0s.defaults.options.json)

		# decrypt env.
		elif self.arguments.present('--decrypt-env'):
			env = self.arguments.get('--decrypt-env')
			key = f"{env}/key/".replace("//","/").replace("//","/").replace("//","/")
			input = f"{env}/data.enc.zip".replace("//","/").replace("//","/").replace("//","/")
			output = f"{env}/data/".replace("//","/").replace("//","/").replace("//","/")
			encryption = self.get_encryption(key=key)
			response = encryption.load_private_key()
			dev0s.response.log(response=response, json=dev0s.defaults.options.json)
			if not response.success: sys.exit(1)
			response = encryption.decrypt_directory(input=input, output=output, remove=self.arguments.present("--remove"))
			self.stop(response=response, json=dev0s.defaults.options.json)

		# generate-keys.
		elif self.arguments.present('--generate-keys'):
			encryption = self.get_encryption(check_passphrase=True)
			response = encryption.generate_keys()
			self.stop(response=response, json=dev0s.defaults.options.json)

		# generate-aes.
		elif self.arguments.present('--generate-aes'):
			self.stop(message=f"Generated AES Passphrase: {utils.__generate__(length=int(self.arguments.get('--length', required=False, default=64)), capitalize=True, digits=True)}", json=dev0s.defaults.options.json)

		# generate passphrase.
		elif self.arguments.present('--generate-passphrase'):
			self.stop(message=f"Generated passphrase: {String('').generate(length=length, capitalize=True, digits=True)}", json=dev0s.defaults.options.json)

		# ____________________________________________________________________________________________________
		#
		# SYSTEM.
		#

		# user.
		elif self.arguments.present(['--user']):
			dev0s.defaults.operating_system(supported=["linux"])

			# initialize a user object.
			user = dev0s.system.User(self.arguments.get("--user"))

			# check if the user exists.
			if self.arguments.present(['--check']):
				response = user.check()
				dev0s.response.log(response=response)
				if response.success: print("User existance:",response["exists"])

			# create a user.
			elif self.arguments.present(['--create']):
				response = user.create()
				self.stop(response=response)

			# delete a user.
			elif self.arguments.present(['--delete']):
				response = user.delete()
				self.stop(response=response)

			# set a users password.
			elif self.arguments.present(['--password']):
				password = self.get_password(retrieve=True, message=f"Enter a new password of user [{user.username}]:")
				response = user.set_password(password=password)
				self.stop(response=response)

			# add the user to groups.
			elif self.arguments.present(['--add-groups']):
				groups = self.arguments.get("--delete-groups").split(",")
				response = user.add_groups(groups=groups)
				self.stop(response=response)

			# delete the user from groups.
			elif self.arguments.present(['--delete-groups']):
				groups = self.arguments.get("--delete-groups").split(",")
				response = user.add_groups(groups=groups)
				self.stop(response=response)


			# invalid.
			else:  self.invalid()

		# group.
		elif self.arguments.present(['--group']):
			dev0s.defaults.operating_system(supported=["linux"])

			# initialize a group object.
			group = dev0s.system.Group(self.arguments.get("--group"))

			# check if the group exists.
			if self.arguments.present(['--check']):
				response = group.check()
				dev0s.response.log(response=response)
				if response.success: 
					print("Group existance:",response["exists"])

			# create a group.
			elif self.arguments.present(['--create']):
				response = group.create()
				self.stop(response=response)

			# delete a group.
			elif self.arguments.present(['--delete']):
				response = group.delete()
				self.stop(response=response)

			# list the current users.
			elif self.arguments.present(['--list-users']):
				response = group.list_users()
				self.stop(response=response)
				if response.success: 
					print(f"Users of group {group.name}:",response["users"])

			# add users to the group.
			elif self.arguments.present(['--add-users']):
				users = self.arguments.get("--add-users").split(",")
				response = group.add_users(users=users)
				self.stop(response=response)

			# delete users from the group.
			elif self.arguments.present(['--delete-users']):
				users = self.arguments.get("--delete-users").split(",")
				response = group.delete_users(users=users)
				self.stop(response=response)

			# check if the specified users are enabled and remove all other users.
			elif self.arguments.present(['--force-users']):
				users = self.arguments.get("--force-users").split(",")
				response = group.check_users(users=users)
				self.stop(response=response)


			# invalid.
			else:  self.invalid()

		# free disk space.
		elif self.arguments.present(["--disk-space"]):
			path = self.arguments.get("--disk-space", required=False, default="/")
			fp = FilePath(path)
			if not fp.exists(): 
				print(f"File path [{fp.path}] does not exist.")
				sys.exit(1)
			loader = dev0s.console.Loader(f"Retrieving size of {fp.path} ...")
			info = fp.space()
			loader.stop()
			print(f"{fp.path}\n * total: {info['total']}\n * used: {info['used']}\n * free: {info['free']}")
			
		# size.
		elif self.arguments.present(["--size"]):
			fp = FilePath(self.arguments.get("--size"))
			if not fp.exists(): 
				print(f"File path [{fp.path}] does not exist.")
				sys.exit(1)
			else:
				loader = dev0s.console.Loader(f"Retrieving size of {fp.path} ...")
				size = fp.size()
				loader.stop()
				print(f"{fp.path}\n * size: {size}")

		# ____________________________________________________________________________________________________
		#
		# METRICS
		#

		# network.
		elif self.arguments.present(['--metrics']):

			# ram.
			if self.arguments.present(['--ram']):
				response = dev0s.system.metrics.ram()
				dev0s.response.log(response=response)
				if response.success: 
					del response["message"] ; del response["error"] ; del response["success"]
					print("RAM metrics:")
					print(response.json(indent=2).replace("{","").replace("}","").replace('"',"").replace(',',"").replace('\n    \n',"\n")[1:-4])


			# invalid.
			else:  self.invalid()

		# ____________________________________________________________________________________________________
		#
		# CLEANUP
		#

		# cleanup.
		elif self.arguments.present(['--cleanup', '--clean-up']):

			# ram.
			if self.arguments.present(['--ram']):
				self.stop(response=dev0s.system.cleanup.ram())

			# swap.
			elif self.arguments.present(['--swap']):
				self.stop(response=dev0s.system.cleanup.swap())

			# invalid.
			else:  self.invalid()

		# ____________________________________________________________________________________________________
		#
		# NETWORK
		#

		# network.
		elif self.arguments.present(['--network']):

			# info.
			if self.arguments.present(['--info']):
				response = dev0s.network.info()
				dev0s.response.log(response=response)
				if response.success: 
					del response["message"] ; del response["error"] ; del response["success"]
					print("Network info:")
					print(response.json(indent=3)[:-2].replace('",',"").replace('"',"").replace('\n  ',"\n * ")[2:])


			# invalid.
			else:  self.invalid()

		# firewall.
		elif self.arguments.present(['--firewall']):

			# check os.
			dev0s.defaults.operating_system(supported=["linux"])

			# retrieve the firewall information.
			if self.arguments.present(['--info']):
				response = dev0s.network.firewall.info()
				dev0s.response.log(response=response)
				if response.success: 
					del response["message"] ; del response["error"] ; del response["success"]
					print(response.json(indent=3).replace("{\n","").replace("\n}","").replace('",',"").replace('"',"").replace('\n  ',"\n * "))

			# disable the firewall.
			elif self.arguments.present(['--disable']):
				response = dev0s.network.firewall.disable()
				dev0s.response.log(response=response)

			# enable the firewall.
			elif self.arguments.present(['--enable']):
				response = dev0s.network.firewall.enable()
				dev0s.response.log(response=response)

			# set the default port action.
			elif self.arguments.present(['--set-default']):
				deny = self.arguments.get("--set-default")
				if deny in ["True", "true", True]: deny = True
				else: deny = False
				response = dev0s.network.firewall.set_default(deny=deny)
				dev0s.response.log(response=response)

			# allow a port.
			elif self.arguments.present(['--allow']):
				port = int(self.arguments.get("--allow"))
				response = dev0s.network.firewall.allow(port)
				dev0s.response.log(response=response)

			# deny a port.
			elif self.arguments.present(['--deny']):
				port = int(self.arguments.get("--deny"))
				response = dev0s.network.firewall.deny(port)
				dev0s.response.log(response=response)


			# invalid.
			else:  self.invalid()

		# ____________________________________________________________________________________________________
		#
		# INSTALLATION.
		#

		# install.
		elif self.arguments.present('--install'):
			response = manager.installation.install()
			if response.message != None and "Successfully installed " not in response.message: dev0s.response.log(response=response)

		# uninstall.
		elif self.arguments.present('--uninstall'):
			response = manager.installation.uninstall()
			dev0s.response.log(response=response)

		# reinstall.
		elif self.arguments.present('--reinstall'):
			response = manager.installation.uninstall()
			dev0s.response.log(response=response)
			response = manager.installation.install()
			if response.message != None and "Successfully installed " not in response.message: dev0s.response.log(response=response)

		# link.
		elif self.arguments.present('--link'):
			response = manager.installation.link()
			dev0s.response.log(response=response)

		# unlink.
		elif self.arguments.present('--unlink'):
			response = manager.installation.unlink()
			dev0s.response.log(response=response)

		# ____________________________________________________________________________________________________
		#
		# INVALID.
		#
		else: self.invalid()

		#
	def get_encryption(self, prompt_passphrase=True, check_passphrase=False, key=None):
		# key.
		public_key = self.arguments.get('--public-key', required=False)
		private_key = self.arguments.get('--private-key', required=False)
		if public_key == None and private_key == None:
			if key == None: key = self.arguments.get('--key', required=True, json=dev0s.defaults.options.json)
			public_key = f"{key}/public_key"
			private_key = f"{key}/private_key"
		# passphrase.
		passphrase = None
		if prompt_passphrase:
			passphrase = self.arguments.get('-p', required=False)
			if passphrase == None:
				passphrase = self.arguments.get('--passphrase', required=False)
			if passphrase == None:
				passphrase = getpass.getpass("Enter the key's passphrase (leave blank to use no passphrase):")
				if check_passphrase and passphrase != getpass.getpass("Enter the same passphrase:"):
					print("Error: passphrases do not match.")
					sys.exit(1)
			if passphrase in ["", "none", "null"]: passphrase = None
		# encryption.
		return dev0s.encryption.AsymmetricAES(
			public_key=public_key,
			private_key=private_key,
			passphrase=passphrase,)
# main.
if __name__ == "__main__":
	cli = CLI()
	cli.start()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# _____________________________________________________________________
#
# Shortcuts 
#   * imported by import order.
#   * should be imported as.
# 		from dev0s.shortcuts import *
# 
#

# default pip imports.
import os, sys

# import.
import dev0s

# utils.
utils = dev0s.utils

# exceptions.
exceptions = dev0s.exceptions

# color.
color = dev0s.color
symbol = dev0s.symbol

# console.
Loader = dev0s.console.Loader
ProgressLoader = dev0s.console.ProgressLoader

# files & formats.
Files = dev0s.files.Files
Formats = dev0s.files.Formats
gfp = dev0s.files.gfp
gdate = dev0s.files.gdate

FilePath = dev0s.files.FilePath 
String = dev0s.files.String 
Boolean = dev0s.files.Boolean 
Integer = dev0s.files.Integer 
Date = dev0s.files.Date

File = dev0s.files.File
Directory = dev0s.files.Directory
Zip = dev0s.files.Zip
Image = dev0s.files.Image
Bytes = dev0s.files.Bytes
Dictionary = dev0s.files.Dictionary
Array = dev0s.files.Array

Generate = dev0s.files.Generate
Interval = dev0s.files.Interval
Speed = dev0s.files.Speed
speed = dev0s.files.Speed()


# response.
ResponseObject = dev0s.response.ResponseObject

# objects.
Traceback = dev0s.objects.Traceback
Object = dev0s.objects.Object
Thread = dev0s.objects.Thread
ThreadManager = dev0s.objects.ThreadManager

# class database.
Database = dev0s.database.Database

# requests.
RestAPI = dev0s.requests.RestAPI

# code.
OutputObject = dev0s.code.OutputObject
Spawn = dev0s.code.Spawn
Version = dev0s.code.Version 
Script = dev0s.code.Script
Python = dev0s.code.Python

#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from dev0s.classes.defaults.exceptions import Exceptions
from dev0s.classes.defaults.files import *

# pip imports.
from django.http import JsonResponse
import json as pypi_json
import traceback as _traceback_

# the response manager class.
class Response(object):
	def __init__(self):	

		# docs.
		DOCS = {
			"module":"dev0s.response", 
			"initialized":True,
			"description":[], 
			"chapter":"Response", }

		# set log file.
		#	assign new values with self.log_file.assign(...) to keep it shared across other objects.
		self.log_file = String(str(None))

		# set log level.
		#	assign new values with self.log_level.assign(...) to keep it shared across other objects.
		self.log_level = Integer(0)
		
		# objects.
		self.parameters = Parameters()

		# for imports.
		self.ResponseObject = ResponseObject

		# log the timestamps by default.
		self.log_timestamps = True

	# response functions.
	def success(self,
		# the message (must be param #1).
		message, 
		# additional returnable functions (must be param #2).
		variables={}, 
		# log log level of the message (int).
		log_level=None, 
		# the required log level for when printed to console (leave None to use self.log_level).
		required_log_level=None, 
		# save the error to the logs file.
		save=False,
		# return as a django Jsonresponse.
		django=False,
	):
		response = self.response({
			"success":True,
			"message":message,
			"error":None,
		})
		for key, value in variables.items():
			response[key] = value
		self.log(message=response["message"], log_level=log_level, save=save, required_log_level=required_log_level)
		if django: 
			try:
				response = JsonResponse(response.dict(safe=True), safe=False)
			except AttributeError:
				response = JsonResponse(response, safe=False)
		return response
	def error(self,
		# the error message.
		error="", 
		# log log level of the message (int).
		log_level=None, 
		# the required log level for when printed to console (leave None to use self.log_level).
		required_log_level=None, 
		# save the error to the erros file.
		save=False,
		# return as a django Jsonresponse.
		django=False,
		# raise error for developer traceback.
		traceback=ERROR_TRACEBACK,
	):
		response = self.response({
			"success":False,
			"message":None,
			"error":error,
		})
		self.log(error=response["error"], log_level=log_level, save=save, required_log_level=required_log_level)
		if traceback:
			raise ValueError(response["error"])
		if django: 
			try:
				response = JsonResponse(response.dict(safe=True), safe=False)
			except AttributeError:
				response = JsonResponse(response, safe=False)
		return response
		#

	# log functions.
	def log(self, 
		# option 1:
		# the message (#1 param).
		message=None,
		# option 2:
		# the error.
		error=None,
		# option 3:
		# the response dict (leave message None to use).
		response={},
		# print the response as json.
		json=False,
		# optionals:
		# the active log level.
		log_level=0,
		# the required log level for when printed to console (leave None to use self.log_level).
		required_log_level=None, 
		# save to log file.
		save=False,
		# save errors always (for options 2 & 3 only).
		save_errors=None,
		# the log mode (leave None for default).
		mode=None,
	):
		if mode != None: mode = str(mode).lower().replace("-","").replace("_","")
		if mode not in [None, "warning", "alert"]:
			raise Exceptions.InvalidUsage(f"{self.__traceback__(function='log', parameter='mode')}: Selected an invalid mode [{mode}], options: [None, warning, alert].")
		def fill_message(msg, error=False):
			if mode == None:
				if error:
					msg = f"Error: {msg}"
				else:
					msg = msg
			elif mode == "warning":
				msg = f"&RED&Warning&END&: {msg}"
			elif mode == "alert":
				msg = f"&ORANGE&Alert&END&: {msg}"
			return msg
			#
		msg, _error_ = None, False
		if [message,error,response] == [None,None,{}]:
			raise Exceptions.InvalidUsage(f"{self.__traceback__(function='log')}: Define either parameter [message:str], [error:str] or [response:dict].")
		if response != {}:
			if response["error"] != None: 
				_error_ = True
				msg = f"Error: {response['error']}"
			else: 
				if response.__class__.__name__ in ["Output"]: 
					msg = response.output
				else:
					msg = response["message"]
		elif isinstance(error, (str, String)): 
			msg = fill_message(error, error=True)
		else: 
			msg = fill_message(message, error=False)
		if required_log_level == None: required_log_level = self.log_level
		try:
			required_log_level = int(required_log_level)
		except: 
			required_log_level = 0
		try:
			comparison = log_level != None and log_level >= required_log_level
		except TypeError as e:
			if "not supported between instances of 'dict' and 'int'" in f"{e}":
				raise TypeError(f"You most likely returned a response.error when you meant a response.success, error: {e}")
			else:
				raise TypeError(e)
		if comparison:
			#print(f"{Date().seconds_timestamp} - {color.fill(msg)}")
			if json:
				if response != {}:
					print(response.json())
				elif error != None:
					print(self.error(error))
				else:
					print(self.success(message))
			else:
				if self.log_timestamps:
					print(f"{Date().seconds_timestamp} - {color.fill(msg)}")
				else:
					print(f"{color.fill(msg)}")
		if save: 
			self.log_to_file(msg)
		elif save_errors and _error_:
			self.log_to_file(msg)

		#
	def load_logs(self, format="webserver", options=["webserver", "cli", "array", "string"]):
		if self.log_file == None:
			return self.error("Define the dev0s.response.log_file attribute.")
		if not Files.exists(self.log_file): Files.create(self.log_file)
		try:
			logs = Files.load(self.log_file)
		except Exception as e:
			return self.error(f"Failed to load the logs, error: {e}.")
		if format == "webserver":
			while True:
				if "<!DOCTYPE html>" in logs:
					old = str(logs)
					logs = String(logs).replace_between(["<!DOCTYPE html>", "</html>"], " *** HTML CODE ****")
					if logs == old:
						return self.error("Unable remove the html code from the logs.")
				else: break
			logs = logs.replace("\n", "<br>")
		elif format == "cli":
			a=1
		elif format == "array" or format == list:
			logs = logs.split("\n")
		elif format == "string" or format == str:
			logs = str(logs)
		else: 
			return self.error(f"Invalid format parameter [{format}], valid options: {options}.")

		return self.success("Succesfully loaded the logs.", {"logs":logs})
	def reset_logs(self, format="webserver", options=["webserver", "cli", "array", "string"]):
		Files.save(self.log_file, f"Resetted log file.\n")
		response = self.load_logs(format=format)
		if not response.success:  return self.error(f"Failed to load the logs after restting, error: {response.error}")
		return self.success("Succesfully resetted the logs.", {
			"logs":response.logs,
		})
		
		#
	
	# serialize a variable.
	def serialize(self, 
		# the variable to serialize.
		variable={}, 
		# serialize to json format.
		json=False,
		# serialize all unknown objects to str.
		safe=False,
	):
		if variable.__class__.__name__ in ["str", "String", "NoneType", "bool", "Boolean"]:
			if str(variable) in ["true", "True", "TRUE", True]:
				if json:
					return "true"
				else:
					return True
			elif str(variable) in ["false", "False", "FALSE", False]:
				if json:
					return "false"
				else:
					return False
			elif str(variable) in ["null", "None", "Nan", None]:
				if json:
					return "null"
				else:
					return None
			else:
				integer = None
				try:
					if "." in str(variable):
						integer = float(variable)
					else:
						integer = int(variable)
				except:
					integer = None
				if integer != None:
					return integer
				else:
					try:
						variable = pypi_json.loads(variable)
					except:
						try:
							variable = ast.literal_eval(variable)
						except:
							try:
								variable = pypi_json.loads(String(variable).slice_dict())
							except: 
								return str(variable)
		elif isinstance(variable, ResponseObject):
			variable = variable.dict()
		elif variable.__class__.__name__ in ["OutputObject"]:
			variable = variable.response().dict()
		elif isinstance(variable, (dict,Dictionary, list, Array)):
			variable = variable
		elif (json or safe) and not isinstance(variable, (Dictionary)) and isinstance(variable, object):
			if isinstance(variable, (Date)):
				return f'"{variable.raw()}"'
			elif variable.__class__.__name__ in ["Version", "OutputObject"] or isinstance(variable, (Integer,Boolean,Bytes,Array,File,ResponseObject)):
				return variable.raw()	
			elif variable in [None, "None"]:
				return None
			else:
				return str(variable)
		else:
			if json or safe:
				return str(variable)
			else:
				return variable
		if isinstance(variable, (dict, Dictionary)):
			_variable_ = {}
			for key, value in variable.items():
				_variable_[key] = self.serialize(variable=value, json=json)
			return _variable_
		elif isinstance(variable, (list, Array)):
			_variable_ = []
			for value in variable:
				_variable_.append(self.serialize(variable=value, json=json))
			return _variable_
		else:
			if json or safe:
				return str(variable)
			else:
				return variable
		
		#
	def response(self, 
		# the blank response (dict, str, generator) (#1).
		response={
			"success":False,
			"message":None,
			"error":None,
		},
	):
		# serialize shortcut.
		return self.ResponseObject(response)
		#

	# system functions.
	def log_to_file(self, message, raw=False):

		# init.
		try:
			while True:
				if len(message) > 0 and message[len(message)-1] == "\n": message = message[:-1]
				else: break
			with open(self.log_file, "a") as file:
				if raw:
					file.write(f'{message}\n')
				else:
					file.write(f'{Date().seconds_timestamp} - {message}\n')
			response = self.response()
			response["success"] = True
			response["message"] = "Succesfully logged the message."
		
			# check file size.
			size = FilePath(self.log_file).size(mode="mb", format="integer")
			if size >= 100: self.reset_logs()

			# handler.
			return response

		except Exception as e:
			response = self.response()
			response["error"] = f"Failed to log the message, error: {e}."
			return response
			
		#

	# quote & unqoute dict.
	def quote(self, dictionary):
		return urllib.parse.quote(json.dumps(dictionary))
		#
	def unquote(self, encoded):
		return json.loads(urllib.parse.unquote(encoded))
		#

	#
	
# the parameters manager object class.
class Parameters(object):
	def __init__(self):
		
		# docs.
		DOCS = {
			"module":"dev0s.response.parameters", 
			"initialized":True,
			"description":[], 
			"chapter":"Response", }

		#
	
	# get request parameters.
	def get(self, 
		# the django request (1).
		request=None, 
		# the identifiers (#2).
		#	str instance: return the parameters value.
		#	list instance: return a parameters object & return an error response when a parameter is undefined.
		#	dict instance: return a parameters object & return the parameter's value from the dict as a default when undefined.
		parameters=[], 
		# traceback id.
		traceback=None,
	):
		if request == None:
			raise Exceptions.InvalidUsage("<dev0s.response.paramters.get>: Define parameter: [request].")
		
		# single parameter.
		if isinstance(parameters, (str,String)):

			# get params & format.
			parameters = str(parameters)
			format = None
			if ":" in parameters:
				parameters,format = parameters.split(":")
				while True:
					if " " in format: format = format.replace(" ","")
					else: break

			# get variable.
			if request.method in ["post", "POST"]:
				variable = request.POST.get(parameters)
			else:
				variable = request.GET.get(parameters)
			if variable in ["", None]:
				if traceback != None:
					return variable, _response_.error(f"{traceback}: Define parameter: [{parameters}].")
				else:
					return variable, _response_.error(f"Define parameter: [{parameters}].")

			# handle format.
			if format != None:
				format_options = format.split(",")
				found = False
				for format in format_options:
					if format.lower() in ["str", "string"]: 
						try:
							variable = str(variable)
							found = True
						except: a=1
					elif format.lower() in ["int", "integer"]: 
						try:
							variable = int(variable)
							found = True
						except: a=1
					elif format.lower() in ["bool", "boolean"]: 
						try:
							if variable in ["true", "True", "TRUE", True]: variable = True
							else: variable = False
							found = True
						except: a=1
					elif format.lower() in ["float", "double"]: 
						try:
							variable = float(variable)
							found = True
						except: a=1
					elif format.lower() in ["array", "list", "dict", "dictionary"]: 
						try:
							variable = ast.literal_eval(variable)
							found = True
						except:
							try:
								variable = json.loads(variable)
								found = True
							except:
								a=1
					else:
						raise Exceptions.InvalidUsage(f"Selected format [{format}] is not a valid format option.")
					if found:
						break
				if not found:
					return variable, _response_.error(f"Unable to parse excepted format [{Array(format_options).string(joiner=', ')}] from parameter [{parameters}:{variable}].")

			# normalize.
			if variable == "None": variable = None

			# handler.
			return variable, _response_.success(f"Succesfully retrieved request parameter [{parameters}].", {
				"key":parameters,
				"value":variable,
			})

		# list recursive.
		elif isinstance(parameters, (list, Array)):
			optional = False
			params = ResponseObject()
			for param in parameters:
				param_value, response = self.get(request, param, traceback=traceback)
				param = param.split(":")[0]
				if response["error"] != None: 
					if optional:
						params[param] = None
					else:
						return params, response
				else: 
					params[param] = param_value
			if optional:
				for key in parameters:
					try: params[key]
					except: params[key] = None
			return params, _response_.success(f"Succesfully retrieved {len(params)} request parameter(s).")

		# dict recursive.
		elif isinstance(parameters, (dict, Dictionary, ResponseObject)):
			if isinstance(parameters, (ResponseObject)): parameters = parameters.clean()
			optional = True
			params = ResponseObject()
			for param,default in parameters.items():
				param_value, response = self.get(request, param, traceback=traceback)
				param = param.split(":")[0]
				if response["error"] != None: 
					if optional:
						params[param] = default
					else:
						return params, response
				else: 
					params[param] = param_value
			if optional:
				for key,default in parameters.items():
					try: params[key]
					except: params[key] = default
			return params, _response_.success(f"Succesfully retrieved {len(params)} request parameter(s).")


		# invalid.
		else:
			raise Exceptions.InvalidUsage(f"The parameters parameter must be [str, String, list, Array, dict, Dictionary] not [{dictionary.__class__.__name__}].")
		
		#

	# check parameter's values.
	def check(self, 
		# the parameters (dict) (#1).
		parameters={"parameter":None}, 
		# the recognizer value for when the parameters are supposed to be empty.
		default=None, 
		# the traceback id.
		traceback=None,
	):

		# single.
		if isinstance(parameters, tuple):
			name,parameter = parameters
			if parameter == default: 
				if traceback != None:
					return _response_.error(f"{traceback}: Define parameter [{name}].")
				else:
					return _response_.error(f"Define parameter [{name}].")
			if ":" in name:
				name,formats = name.split(":")
				while True:
					if " " in formats: formats = formats.replace(" ","")
					else: break
				formats = formats.split(",")
				param_format = Formats.get(parameter, serialize=True)
				if param_format not in formats:
					return _response_.error(f"Incorrect parameter [{name}:{param_format}] format, correct format(s): {Array(path=False, array=formats).string(joiner=', ')}.")
			return _response_.success(f"Succesfully checked parameter [{name}].")

		# recursive.
		elif isinstance(parameters, (dict,Dictionary,ResponseObject)):
			for id, value in parameters.items():
				response = self.check(parameters=(id, value), default=default, traceback=traceback)
				if response["error"] != None: return response
			return _response_.success(f"Succesfully checked {len(parameters)} parameter(s).")
		# invalid.
		else: 
			raise Exceptions.InstanceError(f"Parameter [parameters] requires to be a [dict, Dictionary, tuple] not [{parameters.__class__.__name__}].")

		#

	#

# the response & parameters object class.
class ResponseObject(object):
	def __init__(self, 
		#
		# Should be initialized with response.success or response.error.
		#
		# the response attributes (dict or dict in str format).
		attributes={
			"success":False,
			"message":None,
			"error":None,
		},
	):
		# docs.
		DOCS = {
			"module":"ResponseObject", 
			"initialized":False,
			"description":[], 
			"chapter": "Response", }

		# serialize attributes to dict.
		serialized = _response_.serialize(attributes)
		if not isinstance(serialized, (dict, Dictionary)):
			raise Exceptions.InvalidUsage(f"<ResponseObject.attributes>: parameter [attributes] must be a [dict, dict in str format], not [{attributes.__class__.__name__}].")
		self.assign(serialized)
		
		# catch error so you can also init custom ResponseObjects.
		try:

			# clean message & error.
			if self.message != None: self.message = String(self.message).capitalized_word()
			if self.error != None: self.error = String(self.error).capitalized_word()
			while True:
				if self.message != None and len(self.message) >= 1 and self.message[len(self.message)-1] in [" ", ".", ","]:
					self.message = self.message[:-1]
				elif self.error != None and len(self.error) >= 1 and self.error[len(self.error)-1] in [" ", ".", ","]:
					self.error = self.error[:-1]
				elif self.error != None and len(self.error) >= len("Error: ") and self.error[:len("Error: ")] in ["Error: "]:
					self.error = String(self.error[len("Error: "):]).capitalized_word()
				elif self.error != None and len(self.error) >= len("..") and String(self.error).last("..") in [".."]:
					self.error = str(String(self.error).remove_last("."))
				else: break
			
			# add dot.
			if self.message != None  and len(self.message) > 0 and self.message[len(self.message)-1] not in ["!", "?"]:
				self.message += "."
			if self.error != None  and len(self.error) > 0 and self.error[len(self.error)-1] not in ["!", "?"]:
				self.error += "."


			# check error passed as success response. & reversed
			if self.message != None and len(self.message) >= len("Failed") and self.message[:len("Failed")] == "Failed":
				#_traceback_.print_exc() 
				raise ValueError("A success response may not start with (failed ...). You most likely called an response.success return while you meant response.error.")
			if self.error != None and len(self.error) >= len("Success") and self.error[:len("Success")] == "Success":
				#_traceback_.print_exc() 
				raise ValueError("An error response may not start with (success ...). You most likely called an response.error return while you meant response.success.")
		except AttributeError: a=1 
		#
	# clean default values.
	def clean(self, 
		# the clean options, select * for all, options: [traceback].
		options=["*"],
		# serialize to ResponseObject (with serialize False the ResponseObject's values are not updated).
		serialize=True,
	):
		
		# options.
		if options in ["*", "all"]: options = ["*"]
		_options = []
		for i in options: _options.append(i.lower())
		options = _options
		
		# clean.
		attributes = self.dict(json=False)
		remove = ["error", "message", "success"]
		if "*" in options or "traceback" in options: remove += ["_traceback_", "__traceback__", "raw_traceback", "_raw_traceback_", "__raw_traceback__"]
		for i in remove:
			try:
				if serialize:
					del self[str(i)]
				else:
					del attributes[str(i)]
			except: a=1
		if serialize:
			return self
		else:
			return attributes

		#
	# assign dict.
	def assign(self, dictionary):
		if isinstance(dictionary, (dict, Dictionary)):
			for key,value in _response_.serialize(dictionary).items():
				self[key] = value
		elif isinstance(dictionary, (tuple, list, Array)):
			for key,value in dictionary:
				self[key] = _response_.serialize(value)
		else:
			raise Exceptions.InvalidUsage("The dictionary parameter must be a dict or tuple.")
		return self
	# crash the error message.
	def crash(self, error="ValueError", traceback=True, json=False, error_only=False):
		if error_only and self["error"] == None: return None
		if json:
			self.log(error=self["error"], json=json)
			sys.exit(1)
		else:
			#if not traceback:
			#	sys.tracebacklimit = 0
			#else:
			#	sys.tracebacklimit = 1
			if error.lower() in ["valueerror", "value_error"]: raise ValueError(self["error"])
			else: raise Exception(self["error"])
	def unpack(self, 
		# the key / keys / defaults parameter (#1).
		# str instance:
		#   unpack the str key
		# list instance:
		#   unpack all keys in the list.
		# dict instance:
		#   unpack all keys from the dict & when not present return the key's value as default.
		keys,
	):
		defaults_ = {}
		if isinstance(keys, (dict, Files.Dictionary, ResponseObject)):
			if isinstance(keys, dict):
				defaults_ = dict(keys)
				keys = list(keys.keys())
			else:
				defaults_ = keys.dict()
				keys = keys.keys()
		elif isinstance(keys, str):
			keys = [keys]
		unpacked = []
		for key in keys:
			value, set = None, True
			try: 
				value = self[key]
			except KeyError: 
				try:
					value = defaults_[key]
				except KeyError: 
					set = False
			if not set:
				raise Exceptions.UnpackError(f"Dictionary does not contain attribute [{key}].")
			unpacked.append(value)
		return unpacked
	def remove(self, keys=[], values=[], save=False):
		for i in keys:
			try: del self[i]
			except: a=1
		if values != []:
			new = {}
			for k,v in self.iterate():
				if v not in values: new[k] = v
			self.assign(new)
		if save: self.save()
		return self
	# iterations.
	def iterate(self, sorted=False, reversed=False):
		return self.items(reversed=reversed, sorted=sorted)
	def items(self, sorted=False, reversed=False, dictionary=None):
		if dictionary == None:
			dictionary = self.dict()
		if sorted: dictionary = self.sort()
		if reversed: return self.reversed().items()
		else: return dictionary.items()
	def keys(self, sorted=False, reversed=False):
		if sorted:
			return self.sort(self.keys(sorted=False, reversed=reversed))
		if reversed: 
			keys = list(vars(self))
			reversed_keys = []
			c = len(keys)-1
			for _ in range(len(keys)):
				reversed_keys.append(keys[c])
				c -= 1
			return reversed_keys
		else: return list(vars(self))
	def values(self, sorted=False, reversed=False, dictionary=None):
		if dictionary == None: dictionary = self.dict()
		if sorted:
			return self.sort(self.values(sorted=False, reversed=reversed, dictionary=dictionary))
		values = []
		for key, value in self.items(reversed=reversed, dictionary=dictionary):
			values.append(value)
		return values
	def reversed(self, dictionary=None):
		if dictionary == None: dictionary = self.dict()
		reversed_dict = []
		for key in self.keys(reversed=True, dictionary=dictionary):
			reversed_dict[key] = dictionary[key]
		return reversed_dict
	def sort(self, alphabetical=True, ascending=False, reversed=False, dictionary=None):
		if dictionary == None: dictionary = self.dict()
		new = {}
		if alphabetical or ascending:
			_sorted_ = Array(path=False, array=list(dictionary.keys())).sort(alphabetical=alphabetical, ascending=ascending, reversed=reversed)
		else: raise ValueError("Unknown behaviour, alphabetical=False.")
		for key in _sorted_:
			new[Formats.denitialize(key)] = dictionary[Formats.denitialize(key)]
		return new
	# return self as dict.
	def dict(self, sorted=False, reversed=False, json=False, safe=False):
		dictionary = {}
		for key in self.keys():
			dictionary[Formats.denitialize(key)] = self[Formats.denitialize(key)]
		return self.serialize(json=json, sorted=sorted, reversed=reversed, safe=safe, dictionary=dictionary)
	# dump json string.
	def json(self, sorted=False, reversed=False, indent=4, safe=True, dictionary=None, ):
		if dictionary == None: dictionary = self.dict()
		return json.dumps(self.serialize(json=True, sorted=sorted, reversed=reversed, safe=safe, dictionary=dictionary), indent=indent).replace(': "false"', ': false').replace(': "true"', ': true').replace(': "null"', ': null')
	# serialize dict.
	def serialize(self, sorted=False, reversed=False, json=False, safe=False, dictionary=None):
		if dictionary == None: dictionary = self.dict()
		if isinstance(dictionary, Dictionary):
			dictionary = dictionary.dictionary
		if sorted:
			d = {}
			for key, value in self.items(reversed=reversed, dictionary=self.sort(alphabetical=True, dictionary=dictionary)):
				d[key] = value
			dictionary = d
		else:
			d = {}
			for key, value in self.items(reversed=reversed, dictionary=dictionary):
				d[key] = value
			dictionary = d
		return _response_.serialize(variable=dictionary, json=json, safe=safe)
	# support default iteration.
	def __iter__(self):
		return iter(self.keys())
	# support '>=' & '>' operator.
	def __gt__(self, response):
		if not isinstance(directory, self.__class__):
			raise Exceptions.InstanceError(f"Can not compare object {self.__class__} & {response.__class__}.")
		return len(self) > len(response)
	def __ge__(self, response):
		if not isinstance(response, self.__class__):
			raise Exceptions.InstanceError(f"Can not compare object {self.__class__} & {response.__class__}.")
		return len(self) >= len(response)
	# support '<=' & '<' operator.
	def __lt__(self, response):
		if not isinstance(response, self.__class__):
			raise Exceptions.InstanceError(f"Can not compare object {self.__class__} & {response.__class__}.")
		return len(self) < len(response)
	def __le__(self, response):
		if not isinstance(response, self.__class__):
			raise Exceptions.InstanceError(f"Can not compare object {self.__class__} & {response.__class__}.")
		return len(self) <= len(response)
	# support '==' & '!=' operator.
	def __eq__(self, dictionary):
		if isinstance(dictionary, dict):
			return str(self.sort()) == str(Dictionary(dictionary).sort())
		elif isinstance(dictionary, Dictionary):
			return str(self.sort()) == str(dictionary.sort())
		else:
			try:
				return str(self.sort()) == str(dictionary.sort())
			except:
				return False
	def __ne__(self, dictionary):
		if isinstance(dictionary, dict):
			return str(self.sort()) != str(Dictionary(dictionary).sort())
		elif isinstance(dictionary, Dictionary):
			return str(self.sort()) != str(dictionary.sort())
		else:
			try:
				return str(self.sort()) != str(dictionary.sort())
			except:
				return False
	# support 'in' operator.
	def __contains__(self, response):
		keys = self.keys()
		if isinstance(response, (list, Array)):
			for i in response:
				if response in keys:
					return True
			return False
		else:
			return response in keys
	# support item assignment.
	def __setitem__(self, key, value):
		setattr(self, key, value)
	def __getitem__(self, key):
		return getattr(self, key)
	def __delitem__(self, key):
		delattr(self, key)
	# str representable.
	def __str__(self):
		return self.json(indent=4) 
	# bool representable.
	def __bool__(self):
		return self.success
		#
	# content count.
	def __len__(self):
		return len(self.keys())
	# object id.
	def __id__(self):
		return f"({self.instance()}:{str(self)})"
	# object instance.
	def instance(self):
		return "ResponseObject"
	@property
	def __name__(self):
		return self.instance()
	# return raw data.
	def raw(self):
		return self.dict()
	# return response self for OutputObject and other objects that init ResponseObject as self and want it to be converted to response.
	def response(self):
		return self
		#

# initialized objects.
# must be initialized as Response since object class Parameters requires it.
response = Response()
_response_ = response
